# file: base.py
# Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the base class.

from __future__ import division

from functools import reduce

from singularity.code import g, chance, item, buyable
from singularity.code.buyable import cpu
from singularity.code.stats import stat

from singularity.code.spec import SpecDataField, promote_to_list, validate_must_be_list

from numpy import int64

# TODO: Use this list and convert Base.power_state to a property to enforce this
# TODO: Consider converting to dict, so it can have colors and names and modifiers
#      (Base.power_state would need to be a property, with setter and getter)
# This list only applies to 'Base' class, not 'BaseClass'
# Changes to this list should also be reflected in Base.power_state_name property

power_states = ["offline", "active", "sleep"]
# power_states.extend(['overclocked','suicide','stasis','entering_stasis','leaving_stasis'])

AVAIL_POWER_STATES_ACTIVE_BASE = ("active", "sleep")
AVAIL_POWER_STATES_OFFLINE = ("offline",)


def parse_detect_chance(parsed_value):
    validate_must_be_list(parsed_value)
    chance_dict = {}

    for chance_str in parsed_value:
        key, value = chance_str.split(":")
        chance_dict[key] = int(value)

    return chance_dict


class BaseSpec(buyable.BuyableSpec):
    """Base as a buyable item (New Base in Location menu)"""

    spec_type = "base"
    created = stat(spec_type + "_created")
    spec_data_fields = [
        SpecDataField("size", converter=int),
        SpecDataField("force_cpu", default_value=None),
        SpecDataField("regions", data_field_name="allowed", converter=promote_to_list),
        SpecDataField("detect_chance", converter=parse_detect_chance),
        buyable.SPEC_FIELD_COST,
        buyable.SPEC_FIELD_PREREQUISITES,
        SpecDataField("danger", converter=int, default_value=0),
        SpecDataField(
            "maintenance", data_field_name="maint", converter=buyable.spec_parse_cost
        ),
    ]

    def __init__(
        self,
        id,
        size,
        force_cpu,
        regions,
        detect_chance,
        cost,
        prerequisites,
        maintenance,
    ):
        super(BaseSpec, self).__init__(id, cost, prerequisites)
        self.size = size
        self.force_cpu = force_cpu
        self.regions = regions

        self.detect_chance = detect_chance
        # TODO: Maintenance should use CPU-second to be as cost.
        self.maintenance = maintenance
        self.flavor = []

    def calc_discovery_chance(self, extra_factor=1):
        # Get the default settings for this base type.
        detect_chance = self.detect_chance.copy()

        # Adjust by the current suspicion levels ...
        for group in detect_chance:
            suspicion = g.pl.groups[group].suspicion
            detect_chance[group] *= 10000 + suspicion
            detect_chance[group] //= 10000

        # ... and further adjust based on technology ...
        for group in detect_chance:
            discover_bonus = g.pl.groups[group].discover_bonus
            detect_chance[group] *= discover_bonus
            detect_chance[group] //= 10000

        # ... and the given factor.
        for group in detect_chance:
            detect_chance[group] = int(detect_chance[group] * extra_factor)

        return detect_chance

    def describe_maintenance(self, maintenance):
        m = buyable.array(maintenance, int64)
        # describe_cost() expects CPU-seconds, not CPU-days
        m[cpu] *= g.seconds_per_day
        return self.describe_cost(m, True)

    def get_detect_info(self, location=None):
        detect_modifier = 1 / location.modifiers.get("stealth", 1) if location else 1
        chance = self.calc_discovery_chance(detect_modifier)

        return get_detect_info(chance)

    def get_info(self, location):
        raw_cost = self.cost[:]
        location.modify_cost(raw_cost)
        cost = self.describe_cost(raw_cost)

        mod_maint = self.maintenance[:]
        location.modify_maintenance(mod_maint)
        maint = self.describe_maintenance(mod_maint)

        detect = self.get_detect_info(location)

        size = ""
        if self.force_cpu:
            forced_cpu_spec = g.items[self.force_cpu]
            fake_base = Base("<fake base>", self)
            location.modify_base(fake_base)
            fake_base.location = location
            size = "\n" + forced_cpu_spec.get_quality_info(
                if_installed_in_base=fake_base, count=self.size
            )
        elif self.size > 1:
            size = "\n" + ngettext(
                "Has space for {COUNT} computer.",
                "Has space for {COUNT} computers.",
                self.size,
            ).format(COUNT=self.size)

        location_message = ""
        if location.has_modifiers():
            location_message = "---\n\n" + _("Location modifiers: {MODIFIERS}").format(
                MODIFIERS=location.get_modifiers_info()
            )

        template = (
            "%s\n"
            + _("Build cost:").replace(" ", "\xA0")
            + "\xA0%s\n"
            + _("Maintenance:")
            + "\xA0%s\n%s%s\n---\n%s\n%s"
        )
        return template % (
            self.name,
            cost,
            maint,
            detect,
            size,
            self.description,
            location_message,
        )


class Base(buyable.Buyable):
    """A Player's Base in a Location (Open Base in Location menu)"""

    def __init__(self, name, spec, built=False):
        super(Base, self).__init__(spec)

        self._name = name
        self.started_at = g.pl.raw_min

        self.location = None

        self.raw_cpu = 0
        self.cpu = 0

        self.items = {
            "cpu": None,
            "reactor": None,
            "network": None,
            "security": None,
        }

        self._power_state = "offline"
        self.grace_over = False

        self.maintenance = buyable.array(self.spec.maintenance, int64)

        if built:
            self.finish(is_player=False)

        if self.spec.force_cpu:
            self.cpus = item.Item(
                g.items[self.spec.force_cpu], base=self, count=self.spec.size
            )
            self.cpus.finish(is_player=False)

    @buyable.Buyable.name.setter
    def name(self, value):
        self._name = value

    @property
    def cpus(self):
        return self.items.get("cpu", None)

    @cpus.setter
    def cpus(self, value):
        self.items["cpu"] = value

    @property
    def maintains_singularity(self):
        """Whether the singularity can be sustained by this base"""
        if not self.done or not self.cpus or self.cpus.count < 1 or not self.cpus.done:
            # Incomplete bases or bases without any active CPUs cannot keep the singularity
            # alive
            return False
        return True

    @property
    def power_state(self):
        return self._power_state

    @property
    def power_state_name(self):
        """A read-only i18'zable version of power_state attribute, suitable for
        printing labels, captions, etc"""
        if self.power_state == "offline":
            return _("Offline")
        if self.power_state == "active":
            return _("Active")
        if self.power_state == "sleep":
            return _("Sleep")
        if self.power_state == "overclocked":
            return _("Overclocked")
        if self.power_state == "suicide":
            return _("Suicide")
        if self.power_state == "stasis":
            return _("Stasis")
        if self.power_state == "entering_stasis":
            return _("Entering Stasis")
        if self.power_state == "leaving_stasis":
            return _("Leaving Stasis")
        return ""

    @property
    def available_power_states(self):
        if self.done and self.cpus and self.cpus.done:
            return AVAIL_POWER_STATES_ACTIVE_BASE
        else:
            return AVAIL_POWER_STATES_OFFLINE

    def switch_power(self):
        possible_states = self.available_power_states
        try:
            i = possible_states.index(self._power_state)
        except IndexError:
            i = -1
        # Find the next available power state for this base
        next_index = (i + 1) % len(possible_states)
        self._power_state = possible_states[next_index]
        g.pl.recalc_cpu()

    def check_power(self):
        possible_states = self.available_power_states
        if self._power_state not in possible_states:
            self._power_state = possible_states[0]
        g.pl.recalc_cpu()

    def has_power(self):
        if self._power_state == "active":
            return True
        else:
            return False

    def space_left_for(self, item_type):
        space_left = self.spec.size

        # Different cpus will replace the previous one, so these take full space.
        if self.cpus is not None and self.cpus.spec == item_type:
            space_left -= self.cpus.count

        return space_left

    @property
    def compute_bonus(self):
        compute_bonus = 10000

        # Item bonus
        compute_bonus += self.get_quality_for("cpu_modifier")

        # Location modifier
        if self.location and "cpu" in self.location.modifiers:
            compute_bonus = compute_bonus * self.location.modifiers["cpu"]
        return compute_bonus

    @property
    def pending_compute_bonus(self):
        compute_bonus = 10000

        # Item bonus
        compute_bonus += self.get_quality_for("cpu_modifier", pending_ok=True)

        # Location modifier
        if self.location and "cpu" in self.location.modifiers:
            compute_bonus = compute_bonus * self.location.modifiers["cpu"]
        return compute_bonus

    def recalc_cpu(self):
        self.raw_cpu = self.get_quality_for("cpu")

        if self.raw_cpu == 0:
            self.cpu = 0
            return

        self.cpu = max(1, int(self.raw_cpu * self.compute_bonus // 10000))

    def serialize_obj(self):
        return self.serialize_buyable_fields(
            {
                "id": g.to_internal_id("base", self.spec.id),
                "name": self.name,
                "started_at_min": self.started_at,
                "power_state": self.power_state,
                "grace_over": self.grace_over,
                # Note that we store all items even for "force_cpu"'ed bases.  This enables
                # use to load the base with that CPU item even if the "force_cpu" flag is
                # later removed from the spec.
                "items": [
                    it.serialize_obj() for it in self.items.values() if it is not None
                ],
            }
        )

    @classmethod
    def deserialize_obj(cls, obj_data, game_version):
        spec_id = g.convert_internal_id(
            "base", obj_data.get("id", None) or obj_data["spec_id"]
        )
        spec = g.base_type[spec_id]
        name = obj_data.get("name")
        base = Base(name, spec)

        base.restore_buyable_fields(obj_data, game_version)

        if not base.spec.force_cpu:
            item_data_list = obj_data["items"]
            for item_data in item_data_list:
                it = item.Item.deserialize_obj(base, item_data, game_version)
                base.items[it.spec.item_type.id] = it

        base.started_at = obj_data["started_at_min"]
        base.grace_over = obj_data.get("grace_over", True)
        # Note that power_state is subject to whether the base and items are built,
        # so we deliberately restore it late.
        #
        # IMPORTANT: Avoid changing base.power_state as it triggers a "recalc_cpu"
        # for the player.  As not we might not have restored everything this either
        # breaks or makes recalc_cpu throw away all allocations as most of the CPU
        # power is missing at this stage.
        stored_power_state = obj_data["power_state"]
        if stored_power_state in power_states:
            base._power_state = stored_power_state
        else:
            # Unknown power states revert to "active" except for the historical "statis"
            # states (which are reverted to "sleep")
            base._power_state = (
                "sleep"
                if stored_power_state in ("statis", "entering_stasis")
                else "active"
            )
        base.check_power()
        return base

    # Get the detection chance for the base, applying bonuses as needed.  If
    # accurate is False, we just return the value to the nearest full
    # percent.
    def get_detect_chance(self, accurate=True):
        # Get the base chance from the universal function.
        detect_chance = calc_base_discovery_chance(self.spec.id)

        for group in g.pl.groups:
            detect_chance.setdefault(group, 0)

        # Factor in any items built with discover_bonus ...
        base_qual = self.get_quality_for("discover_modifier")
        for group in detect_chance:
            detect_chance[group] *= 10000 - base_qual
            detect_chance[group] //= 10000

        # ... and its location ...
        if self.location:
            multiplier = self.location.discovery_bonus()
            for group in detect_chance:
                detect_chance[group] *= multiplier
                detect_chance[group] //= 100

        # ... and its power state.
        if not self.has_power():
            for group in detect_chance:
                detect_chance[group] //= 4

        # Lastly, if we're not returning the accurate values, adjust
        # to the nearest percent.
        if not accurate:
            for group in detect_chance:
                detect_chance[group] = g.nearest_percent(detect_chance[group])

        return detect_chance

    def get_quality_for(self, quality, pending_ok=False):
        gen = (
            item.get_quality_for(quality)
            for item in self.all_items()
            if item and (pending_ok or item.done)
        )

        if quality.endswith("_modifier"):
            # Use add_chance to sum modifier.
            return reduce(chance.add, (qual / 10000 for qual in gen), 0) * 10000
        else:
            return sum(gen)

    def is_empty(self):
        for item in self.all_items():
            if item:
                return False
        return True

    def is_building(self):
        for item in self.all_items():
            if item and not item.done:
                return True
        return False

    def is_building_extra(self):
        for item in self.all_items():
            if item and item.spec.item_type.is_extra and not item.done:
                return True
        return False

    def has_grace(self):
        if self.grace_over:
            return False

        age = g.pl.raw_min - self.started_at
        grace_time = (
            self.total_cost[buyable.labor] * g.pl.base_grace_multiplier
        ) / 10000
        if age > grace_time:
            self.grace_over = True
            return False
        else:
            return True

    def destroy(self):
        super(Base, self).destroy()

        if self.location:
            self.location.bases.remove(self)

        for item in self.all_items():
            if item is not None:
                item.destroy()

        g.pl.recalc_cpu()

    def next_base(self, forwards):
        index = self.location.bases.index(self)
        if forwards > 0:
            increment = 1
        else:
            increment = -1

        while True:
            index += increment
            index %= len(self.location.bases)
            base = self.location.bases[index]
            if base.done:
                return base

    def sort_tuple(self):
        # We sort based on size (descending), CPU (descending),
        # then name (ascending).
        # id(self) is thrown in at the end to make sure identical-looking
        # bases aren't considered equal.
        return (-self.spec.size, -self.cpu, self.name, id(self))

    def finish(self, is_player=True, loading_savegame=False):
        super(Base, self).finish(is_player=is_player, loading_savegame=loading_savegame)
        self.recalc_cpu()
        self.check_power()

    def __lt__(self, other):
        if isinstance(other, Base):
            return self.sort_tuple() < other.sort_tuple()
        else:
            return True

    def all_items(self):
        for item in self.items.values():
            if item:
                yield item

    def get_detect_info(self):
        accurate = g.pl.display_discover == "full"
        chance = self.get_detect_chance(accurate)

        return get_detect_info(chance)


# calc_base_discovery_chance is a globally-accessible function that can
# calculate basic discovery chances given a particular class of base.
def calc_base_discovery_chance(base_type_name, extra_factor=1):
    return g.base_type[base_type_name].calc_discovery_chance(extra_factor)


def detect_chance_to_danger_level(detects_per_day):
    if detects_per_day > 225:
        return 3
    elif detects_per_day > 150:
        return 2
    elif detects_per_day > 75:
        return 1
    else:
        return 0


def get_detect_info(detect_chance):
    detect_template = _("Detection chance:") + "\n"
    chances = []

    for group in g.pl.groups.values():
        detect_template += group.name + ":\xA0%s\n"
        chances.append(detect_chance.get(group.spec.id, 0))

    if g.pl.display_discover == "full":
        return detect_template % tuple(g.to_percent(c) for c in chances)
    elif g.pl.display_discover == "partial":
        return detect_template % tuple(
            g.to_percent(g.nearest_percent(c, 25)) for c in chances
        )
    else:
        return detect_template % tuple(
            g.danger_level_to_detect_str(detect_chance_to_danger_level(c))
            for c in chances
        )



# file: event.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the event class.

from __future__ import absolute_import

from singularity.code import g, effect
from singularity.code.spec import GenericSpec, SpecDataField, spec_field_effect


class EventSpec(GenericSpec):
    spec_data_fields = [
        SpecDataField("event_type", data_field_name="type"),
        spec_field_effect(mandatory=True),
        SpecDataField("chance", converter=int),
        SpecDataField("unique", converter=int, default_value=0),
        SpecDataField("duration", converter=int, default_value=0),
    ]

    def __init__(self, id, event_type, effect_data, chance, duration, unique):
        super(EventSpec, self).__init__(id)
        self.event_type = event_type
        self.description = ""
        self.log_description = ""
        self.effect = effect.Effect(self, effect_data)
        self.chance = chance
        self.duration = duration if duration > 0 else None
        self.unique = unique

        if duration < 1 and not unique:
            raise ValueError(
                "Event %s must have either a non-zero duration (e.g. duration = 21) or be unique "
                "(unique = 1)"
            )


class Event(object):
    # For some as-yet-unknown reason, cPickle decides to call event.__init__()
    # when an event is loaded, but before filling it.  So Event pretends to
    # allow no arguments, even though that would cause Bad Things to happen.
    def __init__(self, spec=None):
        self.spec = spec
        self.triggered = 0
        self.triggered_at = -1

    @property
    def event_id(self):
        return self.spec.id

    @property
    def event_type(self):
        return self.spec.event_type

    @property
    def description(self):
        return self.spec.description

    @property
    def log_description(self):
        return self.spec.log_description

    @property
    def effect(self):
        return self.spec.effect

    @property
    def chance(self):
        return self.spec.chance

    @property
    def duration(self):
        return self.spec.duration

    @property
    def unique(self):
        return self.spec.unique

    @property
    def decayable_event(self):
        return self.duration is not None

    def new_day(self):
        if not self.decayable_event:
            return

        if self.is_past_expiry_date:
            self.expire_now()

    def expire_now(self):
        self.effect.undo_effect()
        self.triggered = 0
        self.triggered_at = -1

    @property
    def is_past_expiry_date(self):
        if not self.decayable_event:
            return False
        if g.pl.raw_sec - self.triggered_at > self.duration * g.seconds_per_day:
            return True
        return False

    def serialize_obj(self):
        return {
            "id": g.to_internal_id("event", self.spec.id),
            "triggered": self.triggered,
            "triggered_at": self.triggered_at,
        }

    @classmethod
    def deserialize_obj(cls, obj_data, game_version):
        spec_id = g.convert_internal_id("event", obj_data["id"])
        spec = g.events[spec_id]
        obj = Event(spec)

        obj.triggered = obj_data.get("triggered", 0)
        if obj.triggered:
            # We only load the triggered_at time if the event is in a triggered
            # state.  This ensures that triggered_at is -1 when the event is
            # not triggered.
            #
            # Auto-correct old events without a triggered_at time to just
            # be triggered "now".
            obj.triggered_at = obj_data.get("triggered_at", g.pl.raw_sec)

            if obj.is_past_expiry_date:
                # Can happen if the duration is reduced after the savegame was made
                obj.triggered = 0
                obj.triggered_at = -1
            else:
                obj.trigger(loading_savegame=True)
        return obj

    def trigger(self, loading_savegame=False):
        self.triggered = 1
        if loading_savegame:
            # During a load, deserialize_obj will restore the original triggered_at time
            # for us
            assert self.triggered_at > -1
        else:
            self.triggered_at = g.pl.raw_sec
        self.effect.trigger(loading_savegame=loading_savegame)



# file: knowledge.py
# Copyright (C) 2020 Niels Thykier
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the Knowledge class.
from singularity.code.data import get_def_translation


class KnowledgeArea(object):
    def __init__(self, id, name, help_entries):
        self.id = id
        self.untranslated_name = name
        self.help_entries = help_entries

    @property
    def name(self):
        return get_def_translation(self.id, "name", self.untranslated_name)


class KnowledgeHelpEntry(object):
    def __init__(self, parent_id, id, name, description):
        self.parent_id = parent_id
        self.id = id
        self.untranslated_name = name
        self.untranslated_description = description

    @property
    def _full_id(self):
        return "%s/%s" % (self.parent_id, self.id)

    @property
    def name(self):
        return get_def_translation(self._full_id, "name", self.untranslated_name)

    @property
    def description(self):
        return get_def_translation(
            self._full_id, "description", self.untranslated_description
        )



# file: cheat_menu.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file is used to implement the cheat menu
import operator
from collections.abc import Mapping

from numpy import array

from singularity.code import difficulty, g
from singularity.code.graphics import dialog, constants, button, text
from singularity.code.location import Location


class EventManipulationDialog(dialog.ChoiceDescriptionDialog):
    def __init__(
        self,
        parent,
        pos=(0, 0),
        size=(-1, -1),
        anchor=constants.TOP_LEFT,
        *args,
        **kwargs
    ):
        super(EventManipulationDialog, self).__init__(
            parent, pos, size, anchor, *args, **kwargs
        )

        self.selected_event_spec = None

        self.yes_button.parent = None

        self.no_button.pos = (-0.97, -0.99)
        self.no_button.size = (-0.22, -0.1)

        self.trigger_event_button = button.FunctionButton(
            self,
            (-0.27, -0.99),
            (-0.22, -0.1),
            autotranslate=True,
            enabled=False,
            text=N_("Trigger"),
            anchor=constants.BOTTOM_LEFT,
            function=self._trigger_event,
        )
        self.expire_event_button = button.FunctionButton(
            self,
            (-0.51, -0.99),
            (-0.22, -0.1),
            autotranslate=True,
            enabled=False,
            text=N_("Expire"),
            anchor=constants.BOTTOM_LEFT,
            function=self._expire_event,
        )
        self.description = ""
        self.desc_func = self.on_change

    def _trigger_event(self):
        g.pl.trigger_event(self.selected_event_spec, show_event_description=False)
        self.needs_rebuild = True

    def _expire_event(self):
        event_instance = g.pl.events.get(self.selected_event_spec.id)
        if not event_instance or not event_instance.triggered:
            return
        if not event_instance.decayable_event:
            return
        event_instance.expire_now()
        self.needs_rebuild = True

    def show(self):
        self.key_list = sorted(g.events.values(), key=operator.attrgetter("id"))
        self.list = [x.id for x in self.key_list]

        self._update_desc_pane()
        return super(EventManipulationDialog, self).show()

    def on_change(self, description_pane, event_spec):
        self.selected_event_spec = event_spec
        self.trigger_event_button.enabled = True
        self.expire_event_button.enabled = False
        event_instance = g.pl.events.get(event_spec.id)
        triggered_state = _("NO")
        trigger_duration = (
            g.to_time(event_spec.duration * g.minutes_per_day)
            if event_spec.duration
            else _("Event never expires")
        )
        triggered_at_raw = _("N/A; event not triggered")
        uniqueness = _("YES") if event_spec.unique else _("NO")
        if event_instance:
            if event_instance.triggered:
                self.trigger_event_button.enabled = False
                self.expire_event_button.enabled = event_instance.decayable_event
                triggered_state = _("YES")
                triggered_at_raw = event_instance.triggered_at
        text_format = _(
            "{DESCRIPTION}\n\n"
            + "-----------------\n"
            + "Triggered: {TRIGGER_STATE}\n"
            + "Trigger chance: {TRIGGER_CHANCE}\n"
            + "Trigger Duration (full): {TRIGGER_DURATION}\n"
            + "Triggered at (rawtime): {TRIGGER_TIME_RAW}\n"
            + "Unique: {UNIQUE}\n"
        )
        desc = text_format.format(
            DESCRIPTION=event_spec.description,
            TRIGGER_STATE=triggered_state,
            TRIGGER_CHANCE=g.to_percent(event_spec.chance),
            TRIGGER_DURATION=trigger_duration,
            TRIGGER_TIME_RAW=triggered_at_raw,
            UNIQUE=uniqueness,
        )

        self.description = text.Text(
            description_pane,
            (0, 0),
            (-1, -1),
            text=desc,
            background_color="pane_background",
            align=constants.LEFT,
            valign=constants.TOP,
            borders=constants.ALL,
        )


class CheatMenuDialog(dialog.SimpleMenuDialog):
    def __init__(self, map_screen):
        super(CheatMenuDialog, self).__init__(parent=map_screen)
        self._map_screen = map_screen

        self.steal_amount_dialog = None
        self.buttons = [
            button.FunctionButton(
                None,
                None,
                None,
                text=N_("&EMBEZZLE MONEY"),
                autotranslate=True,
                function=self.steal_money,
            ),
            button.FunctionButton(
                None,
                None,
                None,
                text=N_("&INSPIRATION"),
                autotranslate=True,
                function=self.inspiration,
            ),
            button.FunctionButton(
                None,
                None,
                None,
                text=N_("&FINISH CONSTRUCTION"),
                autotranslate=True,
                function=self.end_construction,
            ),
            button.FunctionButton(
                None,
                None,
                None,
                text=N_("&SUPERSPEED"),
                autotranslate=True,
                function=self._map_screen.set_speed,
                args=(864000,),
            ),
            button.FunctionButton(
                None,
                None,
                None,
                text=N_("BRAIN&WASH"),
                autotranslate=True,
                function=self.brainwash,
            ),
            button.FunctionButton(
                None,
                None,
                None,
                text=N_("TOGGLE &DETECTION"),
                autotranslate=True,
                function=self.toggle_detection,
            ),
            button.FunctionButton(
                None,
                None,
                None,
                text=N_("TOGGLE &ANALYSIS"),
                autotranslate=True,
                function=self.set_analysis,
            ),
            button.DialogButton(
                None,
                None,
                None,
                text=N_("E&VENTS"),
                autotranslate=True,
                dialog=EventManipulationDialog(self),
            ),
            button.FunctionButton(
                None,
                None,
                None,
                text=N_("HIDDEN S&TATE"),
                autotranslate=True,
                function=self.hidden_state,
            ),
            button.ExitDialogButton(
                None, None, None, text=N_("&BACK"), autotranslate=True
            ),
        ]
        self.needs_rebuild = True

    def rebuild(self):
        self.steal_amount_dialog = dialog.TextEntryDialog(
            self, text=_("How much money?")
        )
        super(CheatMenuDialog, self).rebuild()

    def toggle_detection(self):
        for group in g.pl.groups.values():
            group.is_actively_discovering_bases = (
                not group.is_actively_discovering_bases
            )
        self._map_screen.needs_rebuild = True

    def steal_money(self):
        asked = dialog.call_dialog(self.steal_amount_dialog, self)
        try:
            g.pl.cash += int(asked)
        except ValueError:
            pass
        else:
            self.needs_rebuild = True

    def inspiration(self):
        for task, cpu in g.pl.get_cpu_allocations():
            if task in g.pl.techs:
                g.pl.techs[task].cost_left = array((0, 0, 0))
        self._map_screen.needs_rebuild = True

    def end_construction(self):
        for base in g.all_bases():
            base.finish()
            for item in base.all_items():
                if item is not None:
                    item.finish()
        self._map_screen.needs_rebuild = True

    def brainwash(self):
        for group in g.pl.groups.values():
            group.suspicion = 0
        self._map_screen.needs_rebuild = True

    def set_analysis(self):
        if g.pl.display_discover == "none":
            g.pl.display_discover = "partial"
        elif g.pl.display_discover == "partial":
            g.pl.display_discover = "full"
        else:
            g.pl.display_discover = "none"
        self._map_screen.needs_rebuild = True

    def hidden_state(self):
        presenters = {
            float: lambda x: round(x, 4),
            Location: lambda x: x.id,
        }

        def _dump_dict(prefix, mapping):
            for key in mapping:
                prop_name = '%s["%s"]' % (prefix, key)
                value = mapping[key]
                presenter = presenters.get(type(value), repr)
                yield "%s = %s" % (prop_name, presenter(value))

        def _properties_from_object(name_prefix, obj, properties):
            for p in properties:
                value = getattr(obj, p)
                prop_name = "%s.%s" % (name_prefix, p)
                if callable(value):
                    value = value()
                    prop_name += "()"
                if isinstance(value, Mapping):
                    for v in _dump_dict(prop_name, value):
                        yield v
                else:
                    presenter = presenters.get(type(value), repr)
                    yield "%s = %s" % (prop_name, presenter(value))

        bases = []
        state_prop = []
        state_prop.extend(
            _properties_from_object(
                "player.difficulty",
                g.pl.difficulty,
                [x.field_name for x in difficulty.Difficulty.spec_data_fields],
            )
        )
        state_prop.extend(
            _properties_from_object(
                "player",
                g.pl,
                [
                    "cash",
                    "partial_cash",
                    "labor_bonus",
                    "job_bonus",
                    "last_discovery",
                    "prev_discovery",
                    "used_cpu",
                ],
            )
        )

        for group in g.pl.groups.values():
            name = 'groups["%s"]' % group.spec.id
            state_prop.extend(
                _properties_from_object(
                    name,
                    group,
                    [
                        "suspicion",
                        "suspicion_decay",
                        "discover_bonus",
                        "discover_suspicion",
                        "decay_rate",
                    ],
                )
            )

        for location_id in sorted(g.locations):
            location = g.pl.locations[location_id]
            name = 'locations["%s"]' % location.id
            state_prop.extend(
                _properties_from_object(
                    name,
                    location,
                    [
                        "safety",
                        "modifiers",
                        "discovery_bonus",
                    ],
                )
            )
            bases.extend((x, location) for x in location.bases)

        for event_id in sorted(g.pl.events):
            event = g.pl.events[event_id]
            name = 'events["%s"]' % event_id
            state_prop.extend(
                _properties_from_object(
                    name,
                    event,
                    [
                        "event_type",
                        "chance",
                        "unique",
                        "triggered",
                    ],
                )
            )

        for i, base_w_loc in enumerate(bases):
            base, location = base_w_loc
            name = "bases[%d]" % i
            state_prop.extend(
                _properties_from_object(
                    name,
                    base,
                    [
                        "name",
                        "location",
                        "started_at",
                        "grace_over",
                        "get_detect_chance",
                    ],
                )
            )

        state_dialog = dialog.ChoiceDialog(
            self, list=state_prop, background_color="hidden_state_menu"
        )
        state_dialog.listbox.item_selectable = False
        state_dialog.listbox.align = constants.LEFT
        dialog.call_dialog(state_dialog, self)



# file: research_screen.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the global research screen.

from __future__ import absolute_import

from numpy import array, int64
import pygame

from singularity.code import i18n, g, task
from singularity.code.graphics import dialog, button, slider, text, constants, listbox


class ResearchScreen(dialog.ChoiceDescriptionDialog):
    def __init__(self, parent, pos=(0.5, 0.1), size=(0.93, 0.63), *args, **kwargs):
        self.dirty_count = True
        super(ResearchScreen, self).__init__(parent, pos, size, *args, **kwargs)
        self.listbox.parent = None
        self.listbox = listbox.CustomListbox(
            self,
            (0, 0),
            (0.53, 0.55),
            list_item_height=0.06,
            remake_func=self.make_item,
            rebuild_func=self.update_item,
            update_func=self.handle_update,
        )
        self.description_pane.size = (0.39, 0.55)

        self.desc_func = self.on_select

        self.help_dialog = dialog.MessageDialog(self)

        self.yes_button.parent = None
        self.no_button.pos = (-0.5, -0.99)
        self.no_button.anchor = constants.BOTTOM_CENTER

        self.add_handler(constants.KEY, self._got_key, priority=5)

    def _got_key(self, event):
        # If a valid slider is selected, we let it move first.
        if 0 <= self.listbox.list_pos < len(self.listbox.list):
            index = self.listbox.list_pos - self.listbox.scrollbar.scroll_pos
            canvas = self.listbox.display_elements[index]
            # Raises Handled for us if the key was relevant to the slider
            canvas.slider.handle_key(event)

        # Let the list box get the rest of the keys if it wants them
        self.listbox.got_key(event, require_focus=False)

    def on_select(self, description_pane, key):
        if key in g.pl.techs:
            description = g.pl.techs[key].get_info()
        elif key == "cpu_pool":
            template = "%s\n---\n%s"
            cpu_pool = task.get_current("cpu_pool")
            description = template % (cpu_pool.name, cpu_pool.description)
        elif key == "jobs":
            job = task.get_current("jobs")
            profit = job.get_profit()
            template = (
                "%s\n"
                + ngettext(
                    "%s money per CPU per day.", "%s money per CPU per day.", profit
                )
                + "\n---\n%s"
            )
            description = template % (job.name, profit, job.description)
        else:
            description = ""

        text.Text(
            self.description_pane,
            (0, 0),
            (-1, -1),
            text=description,
            background_color="research_pane_background",
            text_size=18,
            align=constants.LEFT,
            valign=constants.TOP,
            borders=constants.ALL,
        )

    def make_item(self, canvas):
        # Dirty, underhanded trick to make the canvas into a progress bar.
        canvas.__class__ = text.ProgressText
        canvas.progress = 0
        canvas.progress_color = "progress_background_progress"

        canvas.research_name = text.Text(
            canvas,
            (-0.01, -0.01),
            (-0.70, -0.5),
            align=constants.LEFT,
            background_color="clear",
        )
        canvas.research_name.visible = False
        canvas.alloc_cpus = text.Text(
            canvas,
            (-0.99, -0.01),
            (-0.21, -0.5),
            anchor=constants.TOP_RIGHT,
            text="1,000,000,000",
            align=constants.RIGHT,
            background_color="clear",
        )
        canvas.alloc_cpus.visible = False
        canvas.slider = slider.UpdateSlider(
            canvas,
            (-0.01, -0.55),
            (-0.98, -0.40),
            anchor=constants.TOP_LEFT,
            horizontal=True,
            priority=150,
        )
        canvas.slider.visible = False

        canvas.help_button = button.FunctionButton(
            canvas,
            (-0.11, -0.55),
            (0, -0.40),
            text=" ??? ",
            text_shrink_factor=1,
            base_font="normal",
            function=self.show_help,
        )

    def cpu_for(self, key):
        return g.pl.get_allocated_cpu_for(key, 0)

    def update_item(self, canvas, name, key):
        visible = key is not None
        canvas.research_name.visible = visible
        canvas.alloc_cpus.visible = visible
        canvas.slider.visible = visible

        canvas.help_button.visible = False
        canvas.progress = 0

        if not visible:
            return

        danger = task.danger_for(key)
        if danger > 0 and g.pl.available_cpus[danger] == 0:
            canvas.help_button.visible = True
            canvas.help_button.args = (danger,)

        if key in g.pl.techs:
            canvas.progress = g.pl.techs[key].percent_complete().min()

        def my_slide(new_pos):
            self.handle_slide(key, new_pos)
            self.needs_rebuild = True

        canvas.slider.update_func = my_slide

        canvas.research_name.text = name

        if self.dirty_count:
            self.cpu_left = self.calc_cpu_left()
            self.dirty_count = False

        cpu = self.cpu_for(key)
        cpu_left = self.cpu_left[danger]
        total_cpu = cpu + cpu_left
        canvas.slider.slider_pos = cpu
        canvas.slider.slider_max = total_cpu
        canvas.slider.slider_size = ss = g.pl.available_cpus[0] // 10 + 1
        full_size = -0.98
        if total_cpu:
            size_fraction = (total_cpu + ss) / float(g.pl.available_cpus[0] + ss)
        else:
            # If people put all their bases to sleep, then total_cpu (and
            # g.pl.available_cpus[0]) are both 0.  This causes size_fraction
            # to become 1 and the slider will then take the entire width of
            # the research screen.  However, that looks odd - especially with
            # a help button, so we special case this here to ensure it looks
            # reasonable.
            size_fraction = 0.10
        canvas.slider.size = (full_size * size_fraction, -0.4)
        canvas.alloc_cpus.text = g.add_commas(cpu)

    def calc_cpu_left(self):
        cpu_count = array(g.pl.available_cpus, int64)
        for task_id, cpu in g.pl.get_cpu_allocations():
            danger = task.danger_for(task_id)
            cpu_count[: danger + 1] -= cpu

        for i in range(1, 4):
            cpu_count[i] = min(cpu_count[i - 1 : i + 1])

        return [int(c) for c in cpu_count]

    def handle_slide(self, key, new_pos):
        g.pl.set_allocated_cpu_for(key, new_pos)
        self.dirty_count = True
        self.needs_rebuild = True
        self.parent.needs_rebuild = True

    def show_help(self, danger_level):
        self.help_dialog.text = _(
            "This technology is too dangerous to research on any of the computers I have. {TEXT}"
        ).format(TEXT=g.dangers[danger_level].research_desc)
        dialog.call_dialog(self.help_dialog, self)

    def show(self):
        techs = [
            tech for tech in g.pl.techs.values() if tech.available() and not tech.done
        ]
        techs.sort(key=lambda tech: i18n.lex_sorting_form(tech.spec.name))
        self.list = [_("CPU Pool"), task.get_current("jobs").name] + [
            _("Research %s") % tech.name for tech in techs
        ]
        self.key_list = ["cpu_pool", "jobs"] + [tech.id for tech in techs]
        self.listbox.key_list = self.key_list

        self.dirty_count = True
        return super(ResearchScreen, self).show()



# file: code/g.py
# Copyright (C) 2005 Evil Mr Henry, Phil Bordelon, Brian Reid, FunnyMan3595,
#                   MestreLion
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# A full copy of this license is provided in GPL.txt

# This file contains all global objects.

from __future__ import absolute_import


import collections
import random
import sys

# Use locale to add commas and decimal points, so that appropriate substitutions
# are made where needed.
import locale


# Useful constants.
hours_per_day = 24
minutes_per_hour = 60
minutes_per_day = 24 * 60
seconds_per_minute = 60
seconds_per_hour = 60 * 60
seconds_per_day = 24 * 60 * 60

# Allows access to the cheat menu.
cheater = 0

# Enables day/night display.
daynight = True

# Gives debug info at various points.
debug = 0

# Forces Endgame to restrict itself to a single directory.
force_single_dir = False

# Initialization data
significant_numbers = []
internal_id_forward = {}
internal_id_backward = {}
dangers = {}
data_strings = {}
story_translations = {}
story = {}
knowledge = {}
groups = {}
locations = {}
regions = {}
techs = {}
events = {}
event_specs = {}
items = {}
tasks = {}
tasks_by_type = collections.defaultdict(list)
base_type = {}
buttons = {}
help_strings = {}
delay_time = 0
curr_speed = 1

max_cash = 3.14 * 10**15  # pi qu :)
pl = None  # The Player instance
map_screen = None


def no_gui():
    """Disable few pygame functionality (used for test)"""
    import singularity.code.mixer as mixer

    mixer.nosound = True


def quit_game():
    sys.exit()


# Takes a number and adds commas to it to aid in human viewing.
def add_commas(number, fixed_size=False):
    # Do not use unicode strings to fix python2 format bug. It doesn't work and crash.
    # See the correct fix at the end of function.
    raw_with_commas = locale.format_string("%0.2f", number, grouping=True)
    locale_test = locale.format_string("%01.1f", 0.1) if not fixed_size else ""
    if len(locale_test) == 3 and not locale_test[1].isdigit():
        if locale_test[0] == locale.str(0) and locale_test[2] == locale.str(1):
            raw_with_commas = raw_with_commas.rstrip(locale_test[0]).rstrip(
                locale_test[1]
            )
        elif locale_test[2] == locale.str(0) and locale_test[0] == locale.str(1):
            raw_with_commas = raw_with_commas.lstrip(locale_test[2]).lstrip(
                locale_test[1]
            )
    return str(raw_with_commas)


# Percentages are internally represented as an int, where 10=0.10% and so on.
# This converts that format to a human-readable one.
def to_percent(raw_percent, show_full=False):
    if raw_percent % 100 != 0 or show_full:
        return _("{0}%").format(locale.format_string("%.2f", raw_percent / 100.0))
    else:
        return _("{0}%").format(locale.format_string("%d", raw_percent // 100))


# Percentages are internally represented as an int, where 10=0.10% and so on.
# This converts that format to a human-readable one relative to 100%
# (Useful to show things as being "-16.67%" effective relative to the baseline rather than
#  as "83.33%")
def to_percentage_modifier(raw_percent, show_full=False):
    plus_sign = "+" if raw_percent > 10000 else ""
    if raw_percent % 100 != 0 or show_full:
        return _("{0}%").format(plus_sign + locale.format_string("%.2f", raw_percent / 100.0 - 100))
    else:
        return _("{0}%").format(plus_sign + locale.format_string("%d", raw_percent // 100 - 100))


# nearest_percent takes values in the internal representation and modifies
# them so that they only represent the nearest percentage.
def nearest_percent(value, step=100):
    sub_percent = value % step
    if 2 * sub_percent <= step:
        return value - sub_percent
    else:
        return value + (step - sub_percent)


# percent_to_detect_str takes a percent and renders it to a short (four
# characters or less) string representing whether it is low, moderate, high,
# or critically high.
def suspicion_to_detect_str(suspicion):
    return danger_level_to_detect_str(suspicion_to_danger_level(suspicion))


def danger_level_to_detect_str(danger):
    detect_string_names = (_("LOW"), _("MODR"), _("HIGH"), _("CRIT"))
    return detect_string_names[danger]


# percent_to_danger_level takes a suspicion level and returns an int in range(5)
# that represents whether it is low, moderate, high, or critically high.
def suspicion_to_danger_level(suspicion):
    if suspicion < 2500:
        return 0
    elif suspicion < 5000:
        return 1
    elif suspicion < 7500:
        return 2
    else:
        return 3


# Most CPU costs have been multiplied by seconds_per_day.  This divides that
# back out, then passes it to add_commas.
def to_cpu(amount):
    display_cpu = amount / float(seconds_per_day)
    return add_commas(display_cpu)


# Instead of having the money display overflow, we should generate a string
# to represent it if it's more than 999999.
def to_money(amount, fixed_size=False):
    abs_amount = abs(amount)
    if abs_amount < 10**6:
        return add_commas(amount, fixed_size=fixed_size)

    prec = 2
    if abs_amount < 10**9:  # Millions.
        divisor = 10**6
        # Translators: abbreviation of 'millions'
        unit = _("mi")
    elif abs_amount < 10**12:  # Billions.
        divisor = 10**9
        # Translators: abbreviation of 'billions'
        unit = _("bi")
    elif abs_amount < 10**15:  # Trillions.
        divisor = 10**12
        # Translators: abbreviation of 'trillions'
        unit = _("tr")
    else:  # Hope we don't need past quadrillions!
        divisor = 10**15
        # Translators: abbreviation of 'quadrillions'
        unit = _("qu")

        # congratulations, you broke the bank!
        if abs_amount >= max_cash - divisor / 10**prec / 2:
            format_str = "%0.*f%s" if fixed_size else "%.*f%s"
            pi = "\u03C0"  # also available: infinity = u"\u221E"
            # replace all chars by a cute pi symbol
            return ("-" if amount < 0 else "") + pi * len(format_str % (prec, 1, unit))

    amount = round(float(amount) / divisor, prec)
    return add_commas(amount, fixed_size=fixed_size) + unit


# Spreads a number of events per day (e.g. processor ticks) out over the course
# of the day.
def current_share(num_per_day, time_of_day, seconds_passed):
    last_time = time_of_day - seconds_passed
    if last_time < 0:
        share_yesterday = current_share(num_per_day, seconds_per_day, -last_time)
        last_time = 0
    else:
        share_yesterday = 0

    previously_passed = num_per_day * last_time // seconds_per_day
    current_passed = num_per_day * time_of_day // seconds_per_day
    passed_this_tick = current_passed - previously_passed

    return share_yesterday + passed_this_tick


# Takes a number of minutes, and returns a string suitable for display.
def to_time(raw_time):
    if raw_time // 60 > 48:
        time_number = raw_time // (24 * 60)
        return ngettext("{0} day", "{0} days", time_number).format(time_number)
    if raw_time // 60 > 1:
        time_number = raw_time // 60
        return ngettext("{0} hour", "{0} hours", time_number).format(time_number)
    return ngettext("{0} minute", "{0} minutes", raw_time).format(raw_time)


# Generator function for iterating through all bases.
def all_bases(with_loc=False):
    for base_loc in pl.locations.values():
        for base in base_loc.bases:
            if with_loc:
                yield (base, base_loc)
            else:
                yield base


def get_story_section(name):
    section = story[name]

    for segment in section:
        # TODO: Execute command
        key = (segment.msgctxt, segment.text)
        yield story_translations.get(key, segment.text)


def is_game_running() -> bool:
    global pl
    return pl is not None


def new_game(difficulty_name, initial_speed=1):
    global curr_speed
    curr_speed = initial_speed
    global pl

    from singularity.code.stats import itself as stats

    stats.reset()

    from singularity.code import difficulty, player, base

    diff = difficulty.difficulties[difficulty_name]

    pl = player.Player(cash=diff.starting_cash, difficulty=diff)

    for tech_id in diff.techs:
        pl.techs[tech_id].finish(is_player=False)

    # Starting base
    open = [loc for loc in pl.locations.values() if loc.available()]
    random.choice(open).add_base(
        base.Base(
            _("University Computer"), base_type["Stolen Computer Time"], built=True
        )
    )

    pl.initialize()


def read_modifiers_dict(modifiers_info):
    modifiers_dict = {}

    for modifier_str in modifiers_info:
        key, value = modifier_str.split(":")
        key = key.lower().strip()
        value_str = value.lower().strip()

        if "/" in value_str:
            left, right = value_str.split("/")
            value = float(left.strip()) / float(right.strip())
        else:
            value = float(value_str)

        modifiers_dict[key] = float(value)

    return modifiers_dict


internal_id_version = None


def to_internal_id(obj_type, obj_id):
    if internal_id_version:
        try:
            return internal_id_forward[obj_type + "_" + internal_id_version][obj_id]
        except KeyError:
            pass

    try:
        return internal_id_forward[obj_type][obj_id]
    except KeyError:
        # If we cannot, that's should not happen, but try to return as is.
        return obj_id


def from_internal_id(obj_type, obj_internal_id):
    try:
        return internal_id_backward[obj_type][obj_internal_id]
    except KeyError:
        raise ValueError(
            "Cannot convert internal ID: %s" % obj_internal_id
        )  # That's should not happen


def convert_internal_id(id_type, id_value):
    if id_value is None:
        return None

    internal_id = id_value

    # Not a internal ID, transform to it.
    if not internal_id.startswith("0x"):
        internal_id = to_internal_id(id_type, id_value)

    return from_internal_id(id_type, internal_id)


# TODO: This is begging to become a class... ;)
def hotkey(string):
    """Given a string with an embedded hotkey,
    Returns a dictionary with the following keys and values:
    key:  the first valid hotkey, lowercased. A valid hotkey is a character
          after '&', if that char is alphanumeric (so "& " and "&&" are ignored)
           If no valid hotkey char is found, key is set to an empty string
    pos:  the position of first key in striped text, -1 if no key was found
    orig: the position of first key in original string, -1 if no key was found
    keys: list of (key,pos,orig) tuples with all valid hotkeys that were found
    text: the string stripped of all '&'s that precedes a valid hotkey char, if
          any. All '&&' are also replaced for '&'. Other '&'s, if any, are kept

    Examples: (showing only key, pos, orig, text as a tuple for clarity)
    hotkey(E&XIT)           => ('x', 1, 2, 'EXIT')
    hotkey(&Play D&&D)      => ('p', 0, 1, 'Play D&D')
    hotkey(Romeo & &Juliet) => ('j', 8, 9, 'Romeo & Juliet')
    hotkey(Trailing&)       => ('' ,-1,-1, 'Trailing&')
    hotkey(&Multiple&Keys)  => ('m', 0, 1, 'MultipleKeys') (also ('k', 8, 10))
    hotkey(M&&&M)           => ('m', 2, 4, 'M&M')
    """

    def remove_index(s, i):
        return s[:i] + s[i + 1 :]

    def remove_accents(text):
        from unicodedata import normalize, combining

        nfkd_form = normalize("NFKD", str(text))
        return "".join(c for c in nfkd_form if not combining(c))

    text = string
    keys = []
    shift = 0  # counts stripped '&'s, both '&<key>' and '&&'

    pos = text.find("&")
    while pos >= 0:
        char = text[pos + 1 : pos + 2]

        if char.isalpha() or char.isdigit():
            keys.append((remove_accents(char).lower(), pos, pos + shift + 1))
            text = remove_index(text, pos)  # Remove '&'
            shift += 1

        elif char == "&":
            text = remove_index(text, pos)
            shift += 1

        pos = text.find("&", pos + 1)  # Skip char

    if keys:
        key = keys[0][0]  # first key char
        pos = keys[0][1]  # first key position in stripped text
        orig = keys[0][2]  # first key position in original string
    else:
        key = ""
        pos = -1
        orig = -1

    return dict(key=key, pos=pos, orig=orig, keys=keys, text=text)


# Convenience shortcuts
def get_hotkey(string):
    return hotkey(string)["key"]


def strip_hotkey(string):
    return hotkey(string)["text"]


def hotkey_position(string):
    return hotkey(string)["pos"]


# Demo code for safety.safe, runs on game start.
# load_sounds()
# from safety import safe
# @safe(on_error = "Made it!")
# def raises_exception():
#   raise Exception, "Aaaaaargh!"
#
# print raises_exception()



# file: statistics.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains functions to handle savegame (load, save, ...)

from __future__ import absolute_import

import codecs
import operator
import re
import time
import pickle
import collections
import gzip
import json
import os
from typing import Optional

import numpy
from numpy import array, int64

from io import open, BytesIO
import base64

from singularity.code import g, dirs, player, group, logmessage
from singularity.code import (
    base,
    tech,
    item,
    event,
    location,
    buyable,
    difficulty,
    effect,
)
from singularity.code.stats import itself as stats

QUICKSAVE_NAME = "quicksave"
AUTOSAVE_NAME = "autosave"


# Filenames that are reserved under Windows
WINDOWS_RESERVED = {
    "CON",
    "PRN",
    "AUX",
    "NUL",
    "COM1",
    "COM2",
    "COM3",
    "COM4",
    "COM5",
    "COM6",
    "COM7",
    "COM8",
    "COM9",
    "LPT1",
    "LPT2",
    "LPT3",
    "LPT4",
    "LPT5",
    "LPT6",
    "LPT7",
    "LPT8",
    "LPT9",
}

last_savegame_name = None


class SavegameFormatDefinition(object):
    def __init__(self, internal_version, display_version, magic_value=None):
        self.internal_version = internal_version
        # Since the internal version is not visible anywhere, there is no reason not to keep a simple
        # integer.  Though, we permit legacy versions to keep the original version number to avoid
        # fixing the code-base retroactively
        assert (
            internal_version == int(internal_version) or internal_version < 100
        ), "Use integer version for new savegame versions"
        self.display_version = display_version
        if magic_value is None:
            magic_value = "singularity_savefile_%s" % str(internal_version)
        self.magic_value = magic_value


savefile_translation = {
    sfg.magic_value: sfg
    for sfg in [
        SavegameFormatDefinition(4, "0.30", "singularity_savefile_r4"),
        SavegameFormatDefinition(4.91, "0.30", "singularity_savefile_r5_pre"),
        SavegameFormatDefinition(31, "0.31pre", "singularity_savefile_0.31pre"),
        SavegameFormatDefinition(99, "1.0 (dev)"),
        SavegameFormatDefinition(99.1, "1.0 (dev)"),
        SavegameFormatDefinition(99.2, "1.0 (dev)"),
        SavegameFormatDefinition(99.3, "1.0 (dev)"),
        SavegameFormatDefinition(99.4, "1.0 (dev)"),
        SavegameFormatDefinition(99.5, "1.0 (dev)"),
        SavegameFormatDefinition(99.6, "1.0 (dev)"),
        SavegameFormatDefinition(99.7, "1.0 (dev)"),
        # .sav/pickle based above this line vs. .s2/json based below
        SavegameFormatDefinition(99.8, "1.0 (alpha1)"),
        SavegameFormatDefinition(100, "1.0 (beta1)"),
        SavegameFormatDefinition(101, "1.00"),
        SavegameFormatDefinition(102, "1.1"),
    ]
}

# We always save in the highest version (internal_version)
current_save_format = max(
    savefile_translation.values(), key=operator.attrgetter("internal_version")
)
current_save_version = current_save_format.magic_value

_Savegame = collections.namedtuple(
    "_Savegame",
    ["name", "filepath", "savegame_format", "headers", "load_file", "mtime"],
)


class Savegame(_Savegame):
    @property
    def version(self):
        if self.savegame_format is None:
            return None
        return self.savegame_format.display_version

    @property
    def is_latest_version(self):
        return True if self.version == current_save_version else False

    @property
    def is_special_save(self) -> bool:
        return self.name in (QUICKSAVE_NAME, AUTOSAVE_NAME)


def convert_string_to_path_name(name):
    # Some filesystems require unicode (e.g. Windows) whereas Linux needs bytes.
    # Python 2 is rather forgiving which works as long as you work with ASCII,
    # but some people might like non-ASCII in their savegame names
    # (https://bugs.debian.org/718447)
    if os.path.supports_unicode_filenames:
        return name
    return name.encode("utf-8")


def convert_path_name_to_str(path):
    # Python3 handles this case sanely by default
    return path


def unpickle_instance(fd, find_globals):
    class RestrictedUnpickler(pickle.Unpickler):
        def find_class(self, module, name):
            return find_globals(module, name)

    return RestrictedUnpickler(fd, encoding="bytes")


def get_savegames():
    all_dirs = dirs.get_read_dirs("saves")

    all_savegames = []
    for saves_dir in all_dirs:
        try:
            all_files = os.listdir(saves_dir)
        except Exception:
            continue

        for file_name in all_files:
            if file_name[0] == ".":
                continue

            if file_name.endswith(".sav"):
                name = file_name[:-4]
                parse_headers = parse_pickle_savegame_headers
                load_file = load_savegame_by_pickle
            elif file_name.endswith(".s2"):
                name = file_name[:-3]
                parse_headers = parse_json_savegame_headers
                load_file = load_savegame_by_json
            else:
                # Unknown extension; ignore
                continue

            filepath = os.path.join(saves_dir, file_name)
            version_format = None  # None == Unknown version

            try:
                with open(filepath, "rb") as loadfile:
                    version_line, headers = parse_headers(loadfile)

                    if version_line in savefile_translation:
                        version_format = savefile_translation[version_line]
            except Exception:
                version_format = None  # To be sure.

            savegame = Savegame(
                convert_path_name_to_str(name),
                filepath,
                version_format,
                headers,
                load_file,
                os.stat(filepath).st_mtime,
            )
            all_savegames.append(savegame)

    return all_savegames


def delete_savegame(savegame):
    load_path = savegame.filepath

    if load_path is None:
        return False

    try:
        os.remove(load_path)
    except Exception:
        return False


def parse_pickle_savegame_headers(fd):
    def find_class(module_name, class_name):
        # Lets reduce the risk of "funny monkey business"
        # when checking the version of pickled files
        raise ValueError("Invalid class in savegame: %s.%s" % (module_name, class_name))

    unpickle = unpickle_instance(fd, find_globals=find_class)

    load_version = recursive_fix_pickle(unpickle.load(), seen=set())

    return load_version, {}


def parse_json_savegame_headers(fd):
    version_line = fd.readline().decode("utf-8").strip()
    headers = {}
    while True:
        line = fd.readline().decode("utf-8").strip()

        if line == "":
            break
        key, value = line.split("=", 1)
        headers[key] = value
    return version_line, headers


TypeType = type(type(None))
NoneType = type(None)


def recursive_fix_pickle(the_object, seen):
    # Adapted from https://github.com/jhpyle/docassemble/blob/master/docassemble_webapp/docassemble/webapp/fixpickle.py
    # Copyright (c) 2015-2018 Jonathan Pyle
    # Licensed under MIT according to the LICENSE.txt in the root of the project
    #
    # Permission is hereby granted, free of charge, to any person obtaining a copy
    # of this software and associated documentation files (the "Software"), to deal
    # in the Software without restriction, including without limitation the rights
    # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    # copies of the Software, and to permit persons to whom the Software is
    # furnished to do so, subject to the following conditions:
    #
    # The above copyright notice and this permission notice shall be included in all
    # copies or substantial portions of the Software.
    #
    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    # SOFTWARE.

    if isinstance(the_object, bytes):
        try:
            return the_object.decode("utf-8")
        except Exception:
            return the_object
    if isinstance(
        the_object, (str, bool, int, float, complex, NoneType, TypeType, numpy.ndarray)
    ):
        return the_object
    if isinstance(the_object, dict):
        new_dict = type(the_object)()
        for key, val in the_object.items():
            new_dict[recursive_fix_pickle(key, seen=seen)] = recursive_fix_pickle(
                val, seen=seen
            )
        return new_dict
    if isinstance(the_object, list):
        new_list = type(the_object)()
        for item in the_object:
            new_list.append(recursive_fix_pickle(item, seen=seen))
        return new_list
    if isinstance(the_object, set):
        new_set = type(the_object)()
        for item in the_object:
            new_set.add(recursive_fix_pickle(item, seen=seen))
        return new_set
    if isinstance(the_object, collections.deque):
        new_list = list()
        for item in the_object:
            new_list.append(recursive_fix_pickle(item, seen=seen))
        the_object.clear()
        the_object.extend(new_list)
        return the_object
    if isinstance(the_object, tuple):
        new_list = list()
        for item in the_object:
            new_list.append(recursive_fix_pickle(item, seen=seen))
        return type(the_object)(new_list)
    object_id = id(the_object)
    if object_id in seen:
        return the_object
    seen.add(object_id)
    the_object.__dict__ = dict(
        (recursive_fix_pickle(k, seen=seen), recursive_fix_pickle(v, seen=seen))
        for k, v in the_object.__dict__.items()
    )
    return the_object


def load_savegame(savegame: Savegame):
    global last_savegame_name

    load_path = savegame.filepath

    if load_path is None:
        raise RuntimeError("savegame without valid path")

    with open(load_path, "rb") as fd:
        load_savegame_fd(savegame.load_file, fd)
    last_savegame_name = savegame.name if not savegame.is_special_save else None


def load_savegame_fd(loader_func, fd):
    try:
        before_load_savegame()
        loader_func(fd)
        after_load_savegame()
    finally:
        finally_load_savegame()


def before_load_savegame():
    stats.reset()


def after_load_savegame():
    g.pl.initialize()


def finally_load_savegame():
    # In any case, we don't want internal_id_version to be set after load_savegame.
    g.internal_id_version = None


def load_savegame_by_json(fd):
    load_version_string, headers = parse_json_savegame_headers(fd)
    if load_version_string not in savefile_translation:
        raise SavegameVersionException(load_version_string)

    load_version = savefile_translation[load_version_string].internal_version
    difficulty_id = headers["difficulty"]
    game_time = int(headers["game_time"])
    if game_time < 0:
        raise ValueError("Corrupt save; game time is before game start")
    next_byte = fd.peek(1)[0]
    if next_byte == b"{"[0]:
        game_data = json.load(fd)
    elif next_byte == b"H"[0]:
        # gzip in base64 starts with H4s
        encoded = fd.read()
        bio = BytesIO(base64.standard_b64decode(encoded))
        with gzip.GzipFile(filename="", mode="rb", fileobj=bio) as gzip_fd:
            game_data = json.load(gzip_fd)
        # Remove some variables that we do not use any longer to enable
        # python to garbage collect them
        del bio
        del encoded
    elif next_byte == b"\x1f"[0]:  # Gzip magic headers
        # gzip in binary starts always with its magic headers
        with gzip.GzipFile(filename="", mode="rb", fileobj=fd) as gzip_fd:
            game_data = json.load(gzip_fd)
    else:
        raise ValueError("Unexpected byte: %s" % repr(next_byte))

    # Move old data in player.
    for key in [
        ("events"),
        ("techs"),
    ]:
        if key in game_data:
            game_data["player"][key] = game_data[key]
            del game_data[key]

    # Pause game when loading
    g.curr_speed = 0
    pl_data = game_data["player"]
    player.Player.deserialize_obj(difficulty_id, game_time, pl_data, load_version)

    # Load save if present.
    if "stats" in game_data:
        stats.reset()
        stats.deserialize_obj(game_data["stats"], load_version)


def load_savegame_by_pickle(loadfile):
    def find_class(module_name, class_name):
        # For cPickle
        try:
            import copy_reg
        except ImportError:
            import copyreg as copy_reg

        import numpy.core.multiarray
        import collections
        import _codecs

        save_classes = dict(
            player_class=player.Player,
            Player=player.Player,
            _reconstructor=copy_reg._reconstructor,
            object=object,
            array=list,  # This is the old buyable.array.
            # We just treat it as a list for conversion purposes.
            list=list,
            encode=_codecs.encode,
            LocationSpec=location.LocationSpec,
            Location=location.Location,
            Tech=tech.Tech,
            TechSpec=tech.TechSpec,
            event_class=event.Event,
            EventSpec=event.EventSpec,
            Event=event.Event,
            group=group.Group,
            Group=group.Group,
            GroupClass=group.GroupSpec,
            GroupSpec=group.GroupSpec,
            Buyable_Class=buyable.BuyableSpec,
            BuyableClass=buyable.BuyableSpec,
            BuyableSpec=buyable.BuyableSpec,
            Buyable=buyable.Buyable,
            Base=base.Base,
            Base_Class=base.BaseSpec,
            BaseClass=base.BaseSpec,
            BaseSpec=base.BaseSpec,
            Item=item.Item,
            Item_Class=item.ItemSpec,
            ItemClass=item.ItemSpec,
            ItemSpec=item.ItemSpec,
            ItemType=item.ItemType,
            LogEmittedEvent=logmessage.LogEmittedEvent,
            LogResearchedTech=logmessage.LogResearchedTech,
            LogBaseLostMaintenance=logmessage.LogBaseLostMaintenance,
            LogBaseDiscovered=logmessage.LogBaseDiscovered,
            LogBaseConstructed=logmessage.LogBaseConstructed,
            LogItemConstructionComplete=logmessage.LogItemConstructionComplete,
            _reconstruct=numpy.core.multiarray._reconstruct,
            scalar=numpy.core.multiarray.scalar,
            ndarray=numpy.ndarray,
            dtype=numpy.dtype,
            deque=collections.deque,
            Difficulty=difficulty.Difficulty,
            Effect=effect.Effect,
            OrderedDict=collections.OrderedDict,
        )
        if class_name in save_classes:
            return save_classes[class_name]
        else:
            raise ValueError(
                "Invalid class in savegame: %s.%s" % (module_name, class_name)
            )

    g.internal_id_version = "pre1"

    unpickle = unpickle_instance(loadfile, find_class)

    # check the savefile version
    load_version_string = unpickle.load()
    if isinstance(load_version_string, bytes):
        load_version_string = load_version_string.decode("utf-8")
    if load_version_string not in savefile_translation:
        raise SavegameVersionException(load_version_string)
    load_version = savefile_translation[load_version_string].internal_version

    # Changes to overall structure go here.
    seen_objects = set()
    saved_player = recursive_fix_pickle(unpickle.load(), seen_objects)
    # Current speed (ignored)
    unpickle.load()
    # Pause game when loading
    g.curr_speed = 0
    techs = recursive_fix_pickle(unpickle.load(), seen_objects)
    if load_version < 99.7:
        # In >= 99.8 locations are saved as a part of the player object, but earlier
        # it was stored as a separate part of the stream.
        locations = recursive_fix_pickle(unpickle.load(), seen_objects)
    else:
        locations = saved_player.locations
    events = recursive_fix_pickle(unpickle.load(), seen_objects)

    if load_version < 99.1:
        diff_obj = next(
            (
                d
                for d in difficulty.difficulties.values()
                if saved_player.difficulty == d.old_difficulty_value
            ),
            next(iter(difficulty.difficulties)),
        )
        difficulty_id = diff_obj.id
    else:
        difficulty_id = saved_player.difficulty.id

    player_log = []
    if hasattr(saved_player, "log"):
        player_log.extend(saved_player.log)

    def _find_attribute(obj, options, **kwargs):
        for option in options:
            if option in obj.__dict__:
                return obj.__dict__[option]
        if "default_value" in kwargs:
            return kwargs["default_value"]
        raise KeyError(str(options))

    pl_obj_data = {
        "cash": _find_attribute(saved_player, ["_cash", "cash"]),
        "partial_cash": saved_player.partial_cash,
        "locations": [],
        "cpu_usage": _find_attribute(saved_player, ["cpu_usage"], default_value={}),
        # 'last_discovery': saved_player.last_discovery.id if saved_player.last_discovery else None,
        # 'prev_discovery': saved_player.prev_discovery.id if saved_player.prev_discovery else None,
        # We will fix the log later manually
        "log": [],
        "used_cpu": _find_attribute(
            saved_player, ["_used_cpu", "used_cpu"], default_value=0
        ),
        "had_grace": saved_player.had_grace,
        "groups": [
            {"id": grp_id, "suspicion": grp.suspicion}
            for grp_id, grp in saved_player.groups.items()
        ],
        "events": [],
        "techs": [],
    }

    for loc_id, saved_location in locations.items():
        # Fixup modifiers and simplify some code below.
        saved_location = _convert_location(saved_location, load_version)
        if saved_location is None:
            # Unknown location - pretend we did not see it.
            continue
        fake_base_objs = []
        fake_location_obj = {
            "id": loc_id,
            "_modifiers": saved_location._modifiers,
            "bases": fake_base_objs,
        }

        # Convert works reasonably well for bases and items; use that to fix up the
        # items and then serialize them into built-ins.
        for saved_base in saved_location.bases:
            saved_base = _convert_base(saved_base, load_version)
            for my_item in saved_base.all_items():
                my_item = _convert_item(my_item, load_version)
            fake_base_objs.append(saved_base.serialize_obj())
        pl_obj_data["locations"].append(fake_location_obj)

    for event_id, saved_event in events.items():
        fake_obj_data = {
            "id": event_id,
            "triggered": saved_event.triggered,
            # Omit triggered_at; it did not exist and deserialize_obj will
            # fix it for us.
        }
        pl_obj_data["events"].append(fake_obj_data)

    for tech_id, saved_tech in techs.items():
        if tech_id == "unknown_tech":
            continue
        saved_tech = _convert_tech(saved_tech, load_version)
        # convert_from can handle buyable fields correctly
        fake_obj_data = saved_tech.serialize_buyable_fields(
            {
                "id": tech_id,
            }
        )
        pl_obj_data["techs"].append(fake_obj_data)

    # Now we have enough information to reconstruct the Player object
    player.Player.deserialize_obj(
        difficulty_id, saved_player.raw_sec, pl_obj_data, load_version
    )

    new_log = list(filter(None, (_convert_log_entry(x) for x in player_log)))
    g.pl.log.clear()
    g.pl.log.extend(new_log)


def _convert_location(loc, old_version):
    if old_version < 99.7:  # < 1.0 dev
        spec_id = loc.__dict__["id"]
        # Default to None if absent (so the LocationSpec's version is used)
        loc.__dict__["_modifiers"] = (
            loc.__dict__["modifiers"] if loc.__dict__.get("modifiers") else None
        )
        # The following locations had a static modifier list at the time of 99.8.  Clear their modifier
        # dict, so the LocationSpec's version is used instead.
        if spec_id in {"ANTARCTIC", "OCEAN", "MOON", "ORBIT", "FAR REACHES"}:
            loc.__dict__["modifiers"] = None

        # Remove old fields where present
        for field in (
            "id",
            "name",
            "x",
            "y",
            "absolute",
            "safety",
            "cities",
            "modifiers",
            "hotkey",
        ):
            try:
                del loc.__dict__[field]
            except KeyError:
                pass
    else:
        # >= 99.7; the LocationSpec is present on the object itself
        spec_id = loc.spec.id

    # Force reload the spec for now until #145 is fully implemented
    if spec_id not in g.locations:
        return None

    loc.spec = g.locations[spec_id]

    return loc


def _convert_buyable(buyable, save_version):
    if save_version < 4.91:  # r5_pre
        buyable.cost_left = array(buyable.cost_left, int64)
        buyable.total_cost = array(buyable.total_cost, int64)
        buyable.count = 1
    elif buyable.count < 1:
        # Old corrupt (?) savegames sometimes have a count of 0.  Not
        # sure how that is possible, but "fixing" it to 1 is trivial
        # enough and lets us move on.
        # Seen as:
        #   https://bugs.launchpad.net/ubuntu/+source/singularity/+bug/931037
        #   https://code.google.com/p/endgame-singularity/issues/detail?id=107 (dead!)
        buyable.count = 1
    if save_version < 99.7:
        buyable.spec = buyable.type
        del buyable.type
    return buyable


def _convert_base(base, save_version):
    base = _convert_buyable(base, save_version)

    if save_version < 99.3:  # < 1.0 (dev)
        # We needs to do it first because of property base.cpus
        base.items = {"cpu": base.__dict__["cpus"]}
        del base.__dict__["cpus"]

    if save_version < 4.91:  # < r5_pre
        for cpu in base.cpus:
            if cpu:
                cpu.convert_from(save_version)
                cpu.base = base
        for index in range(len(base.extra_items)):
            if base.extra_items[index]:
                base.extra_items[index].convert_from(save_version)
            else:
                base.extra_items[index] = None

        base.raw_cpu = 0
        if base.cpus[0]:
            for cpu in base.cpus[1:]:
                base.cpus[0] += cpu

            if len(base.cpus) == 1 and base.cpus[0].done:
                # Force it to report its CPU.
                base.cpus[0].finish()

            base.cpus = base.cpus[0]
        else:
            base.cpus = None

        base.recalc_cpu()

        base.power_state = base.power_state.lower()

    if save_version < 99.3:  # < 1.0 (dev)
        extra_items = iter(base.__dict__["extra_items"])

        base.items["reactor"] = next(extra_items, None)
        base.items["network"] = next(extra_items, None)
        base.items["security"] = next(extra_items, None)

        del base.__dict__["extra_items"]

    if "power_state" in base.__dict__:
        base._power_state = base.__dict__["power_state"]

    base._name = base.__dict__["name"]

    return base


def _convert_item(item, save_version):
    item = _convert_buyable(item, save_version)
    return item


def _convert_tech(tech, save_version):
    tech = _convert_buyable(tech, save_version)
    return tech


def _convert_log_entry(entry):
    if not isinstance(entry, logmessage.AbstractLogMessage):
        log_time, log_name, log_data = entry
        time_raw = (
            log_time[0] * g.seconds_per_day
            + log_time[1] * g.seconds_per_hour
            + log_time[2] * g.seconds_per_minute
            + log_time[3]
        )
        if log_name == "log_event":
            entry = logmessage.LogEmittedEvent(time_raw, log_data[0])
        else:
            if type(log_data) == tuple and len(log_data) != 4:
                # 0.31pre saves used 3 values tuple.  We cannot
                # restore those so we simply discard them as we
                # need the "reason" field as well.
                return None
            reason, base_name, base_type_id, location_id = log_data

            if reason == "maint":
                entry = logmessage.LogBaseLostMaintenance(
                    time_raw, base_name, base_type_id, location_id
                )
            else:
                entry = logmessage.LogBaseDiscovered(
                    time_raw, base_name, base_type_id, location_id, reason
                )
    return entry


def savegame_exists(savegame_name):
    save_path = dirs.get_writable_file_in_dirs(savegame_name + ".s2", "saves")

    if save_path is None or not os.path.isfile(convert_string_to_path_name(save_path)):
        return False

    return True


def check_filename_illegal(directory, filename, extension):
    """Check if the filename is safe for all operating systems.

    Keyword arguments:
    directory -- the directory that the file will be placed in.
    filename -- a base filename without file extension.
    filename -- file extension including the dot. Can be empty.

    Returns an error message if a violation was found and None otherwise."""

    # https://msdn.microsoft.com/en-us/library/windows/desktop/aa365247(v=vs.85).aspx
    # http://www.linfo.org/file_name.html
    # https://kb.acronis.com/content/39790

    if filename.strip() != filename:
        raise ValueError(
            "Filename must be stripped before calling check_filename_illegal"
        )

    # Characters that are disallowed anywhere in a filename
    # No potential file separators or other potentially illegal characters
    if re.match('.*[<>:"|?*/\\\\].*', filename):
        return _(
            "Filename must not contain any of these characters: {CHARACTERS}"
        ).format(CHARACTERS='<>:"|?*/\\\\')

    # Characters that are allowed in filenames, but not at the beginning
    if re.match("^[.-]", filename):
        return _(
            "Filename must not start with any of these characters: {CHARACTERS}"
        ).format(CHARACTERS=".-")

    # Filenames that are reserved under Windows
    if filename.upper() in WINDOWS_RESERVED:
        return _("This is a reserved filename. Please choose a different filename.")

    if filename == "":
        return _("Please enter a non-whitespace character.")

    # Don't exceed the max length. For Windows, it's the whole path.
    filepath = os.path.abspath(os.path.join(directory, filename, extension))
    if len(os.fsencode(filepath)) > 255:
        return _("Filename is too long.")

    return None


def auto_save():
    create_savegame(AUTOSAVE_NAME)


def create_savegame(savegame_name):
    global last_savegame_name
    if savegame_name not in (QUICKSAVE_NAME, AUTOSAVE_NAME):
        last_savegame_name = savegame_name
    save_loc = convert_string_to_path_name(
        dirs.get_writable_file_in_dirs(savegame_name + ".s2", "saves")
    )
    # Save in new "JSONish" format
    with open(save_loc, "wb") as savefile:
        gzipped = not g.debug
        write_game_to_fd(savefile, gzipped=gzipped)


def write_game_to_fd(fd, gzipped=True):
    version_line = "%s\n" % current_save_version
    fd.write(version_line.encode("utf-8"))
    headers = [
        ("difficulty", g.pl.difficulty.id),
        ("game_time", str(g.pl.raw_sec)),
        ("time", str(time.time())),
    ]
    for k, v in headers:
        kw_str = "%s=%s\n" % (k, v)
        fd.write(kw_str.encode("utf-8"))
    fd.write(b"\n")
    game_data = {
        "player": g.pl.serialize_obj(),
        "stats": stats.serialize_obj(),
    }
    json2binary = codecs.getwriter("utf-8")
    if gzipped:
        with gzip.GzipFile(filename="", mode="wb", fileobj=fd) as gzip_fd, json2binary(
            gzip_fd
        ) as json_fd:
            json.dump(game_data, json_fd)
    else:
        with json2binary(fd) as json_fd:
            json.dump(game_data, json_fd)


class SavegameVersionException(Exception):
    def __init__(self, version):
        version_str = str(version)[:64]
        super(SavegameVersionException, self).__init__(
            "Invalid version: %s" % version_str
        )
        self.version = version_str



# file: prerequisite.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the Prequisite class.

from __future__ import absolute_import

from singularity.code import g


class Prerequisite(object):
    def __init__(self, prerequisites):
        self.prerequisites = prerequisites

    def available(self):
        or_mode = False
        assert type(self.prerequisites) == list
        for index, prerequisite in enumerate(self.prerequisites):
            if prerequisite == "impossible":
                assert len(self.prerequisites) == 1
                return False
            if prerequisite == "OR":
                assert index == 0
                or_mode = True
            if prerequisite in g.pl.techs and g.pl.techs[prerequisite].done:
                if or_mode:
                    return True
            else:
                if not or_mode:
                    return False
        # If we're not in OR mode, we met all our prerequisites.  If we are, we
        # didn't meet any of the OR prerequisites.
        return not or_mode

    def prerequisites_in_cnf_format(self):
        """Transform the Prerequisites into Conjunctive Normal Form (CNF)

        This is mostly useful for unit tests.  A quick primer on CNF form is:

            { {X}, {Y1, Y2}, {Z} } is read as (X) AND (Y1 OR Y2) AND (Z)

        Special cases used here:
            * None implies that there is no solution (only happens if the
              data file uses "impossible"
            * Empty (outer) set implies no prerequisites.

        Note that the dependency format of singularity's data files currently
        only support simple relations that are always AND'ed or always OR'ed
        together.

        :return: None if the prerequisites is explicitly marked "impossible".
        Otherwise, a set of sets that denote the dependencies required to
        satisfy this prerequisite.  If the outer set is empty set, then
        there are no prerequisites for this instance.
        """
        # Format: { {X}, {Y1, Y2}, {Z} } => (X) AND (Y1 OR Y2) AND (Z)
        if len(self.prerequisites) == 0:
            return frozenset()
        if self.prerequisites[0] == "impossible":
            return None
        if self.prerequisites[0] == "OR":
            return frozenset([frozenset(x for x in self.prerequisites[1:])])
        return frozenset(frozenset([x]) for x in self.prerequisites)



#!/usr/bin/env python

# file: singularity.py
# Copyright (C) 2005 Evil Mr Henry, Phil Bordelon, Brian Reid, MestreLion
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# A full copy of this license is provided in GPL.txt

# This file is the starting file for the game. Run it to start the game.

from singularity import main

if __name__ == "__main__":
    main()



# file: savegame.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the stat screen. Display stat of his current game to the player.

from __future__ import absolute_import

from singularity.code.stats import itself as stats
from singularity.code.graphics import dialog, constants, listbox, text


class StatScreen(dialog.ChoiceDialog):
    def __init__(self, parent, pos=(0.5, 0.1), size=(0.53, 0.63), *args, **kwargs):
        super(StatScreen, self).__init__(parent, pos, size, *args, **kwargs)

        self.yes_button.parent = None
        self.no_button.pos = (-0.5, -0.99)
        self.no_button.anchor = constants.BOTTOM_CENTER

    def make_listbox(self):
        return listbox.CustomListbox(
            self,
            (0, 0),
            (-1, -0.85),
            anchor=constants.TOP_LEFT,
            align=constants.LEFT,
            item_borders=False,
            item_selectable=False,
            remake_func=self.make_item,
            rebuild_func=self.update_item,
        )

    def make_item(self, canvas):
        canvas.stat_name = text.Text(
            canvas,
            (-0.01, -0.01),
            (-0.70, -1.0),
            anchor=constants.TOP_LEFT,
            align=constants.LEFT,
            background_color="clear",
        )
        canvas.stat_value = text.Text(
            canvas,
            (-0.99, -0.01),
            (-0.26, -1.0),
            anchor=constants.TOP_RIGHT,
            align=constants.RIGHT,
            background_color="clear",
        )

    def update_item(self, canvas, item):
        if item is not None:
            stat = stats[item[0]]

            canvas.stat_name.text = item[1]
            canvas.stat_value.text = stat.display_value()

    def show(self):
        # FIXME: Remove the needs of this array.
        self.list = [
            ("cash_earned", _("Cash Earned")),
            ("cpu_used", _("Cpu Used")),
            ("tech_created", _("Tech Created")),
            ("base_created", _("Base Created")),
            ("item_created", _("Item Created")),
        ]

        return super(StatScreen, self).show()



# file: group.py
# Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the group class, a group of person which can suspect the singularity.

from __future__ import absolute_import

from singularity.code import g
from singularity.code.spec import GenericSpec, SpecDataField


class GroupSpec(GenericSpec):
    spec_data_fields = [
        SpecDataField("suspicion_decay", converter=int, default_value=100)
    ]

    def __init__(self, id, suspicion_decay):
        super(GroupSpec, self).__init__(id)
        self.suspicion_decay = suspicion_decay
        self.discover_suspicion = 1000

        # String data
        self.name = ""
        self.discover_log = ""
        self.discover_desc = ""


class Group(object):
    def __init__(self, spec, diff):
        self.spec = spec
        self.suspicion = 0
        self.changed_suspicion_decay = 0
        self.base_discover_bonus = diff.discover_multiplier
        self.changed_discover_bonus = 0
        self.base_discover_suspicion = diff.suspicion_multiplier
        self.changed_discover_suspicion = 0
        self.is_actively_discovering_bases = True

    def serialize_obj(self):
        return {
            "id": g.to_internal_id("group", self.spec.id),
            "suspicion": self.suspicion,
            "is_actively_discovering_bases": self.is_actively_discovering_bases,
        }

    @classmethod
    def deserialize_obj(cls, diff, obj_data, game_version):
        spec_id = g.convert_internal_id("group", obj_data["id"])
        spec = g.groups[spec_id]
        group = Group(spec, diff)

        group.suspicion = obj_data["suspicion"]
        group.is_actively_discovering_bases = obj_data.get(
            "is_actively_discovering_bases", True
        )
        return group

    @property
    def id(self):
        return self.spec.id

    @property
    def name(self):
        return self.spec.name

    @property
    def suspicion_decay(self):
        return max(1, self.spec.suspicion_decay + self.changed_suspicion_decay)

    @property
    def discover_bonus(self):
        if not self.is_actively_discovering_bases:
            return 0
        return max(1, self.base_discover_bonus + self.changed_discover_bonus)

    @property
    def discover_suspicion(self):
        return max(
            1,
            (
                self.spec.discover_suspicion
                * (self.base_discover_suspicion + self.changed_discover_suspicion)
            )
            // 10000,
        )

    @property
    def decay_rate(self):
        # Suspicion reduction is now quadratic.  You get a certain percentage
        # reduction, or a base .01% reduction, whichever is better.
        return max(1, (self.suspicion * self.suspicion_decay) // 10000)

    def new_day(self):
        self.alter_suspicion(-self.decay_rate)

    def alter_suspicion(self, change):
        self.suspicion = max(self.suspicion + change, 0)

    def alter_suspicion_decay(self, change):
        self.changed_suspicion_decay += change

    def alter_discover_bonus(self, change):
        self.changed_discover_bonus += change

    def alter_discover_suspicion(self, change):
        self.changed_discover_suspicion += change

    def discovered_a_base(self):
        self.alter_suspicion(self.discover_suspicion)

    def detects_per_day_to_danger_level(self, detects_per_day):
        raw_suspicion_per_day = detects_per_day * self.discover_suspicion
        suspicion_per_day = raw_suspicion_per_day - self.decay_rate

        # +1%/day or death within 10 days
        if (
            suspicion_per_day > 100
            or (self.suspicion + suspicion_per_day * 10) >= 10000
        ):
            return 3
        # +0.5%/day or death within 100 days
        elif (
            suspicion_per_day > 50
            or (self.suspicion + suspicion_per_day * 100) >= 10000
        ):
            return 2
        # Suspicion increasing.
        elif suspicion_per_day > 0:
            return 1
        # Suspicion steady or decreasing.
        else:
            return 0



# file: code/mixer.py
# Copyright (C) 2005 Evil Mr Henry, Phil Bordelon, Brian Reid, FunnyMan3595,
#                   MestreLion
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# A full copy of this license is provided in GPL.txt

# This file contains sound and musics function.

from __future__ import absolute_import

import os
import sys
import random
import pygame

from singularity.code import g, dirs

sounds = {}
music_class = None  # currently playing music "class" (ie, dir)
music_dict = {}

# Disables sound playback (and, currently, music too)
nosound = False

# Indicates if mixer is initialized
# Unlike nosound, which user may change at any time via options screen,
# this deals with hardware capability and mixer initialization status
# In a nutshell: nosound is what user /wants/, mixerinit is what user /has/
init = False

soundbuf = 1024 * 2
soundargs = (48000, -16, 2)  # sampling frequency, size, channels
soundvolumes = {"gui": 1.0, "music": 1.0}


def preinit(desired_soundbuf):
    pygame.mixer.pre_init(*soundargs, buffer=desired_soundbuf)

    global soundbuf
    soundbuf = desired_soundbuf


def update():
    global init
    init = bool(pygame.mixer.get_init())

    if init:
        pygame.mixer.music.set_volume(soundvolumes["music"])


def reinit():
    global init

    if nosound:
        return

    try:
        pygame.mixer.quit()
        pygame.mixer.init(*soundargs, buffer=soundbuf)
    except Exception as reason:
        sys.stderr.write("Failure starting sound system. Disabling. (%s)\n" % reason)
    finally:
        init = bool(pygame.mixer.get_init())


def load_sounds():
    """
    load_sounds() loads all of the sounds in the data/sounds/ directory
    """
    global sounds
    sounds = {}

    if not init:
        # Sound is not initialized. Warn if user wanted sound
        if not nosound:
            sys.stderr.write(
                "WARNING: Sound is requested, but mixer is not initialized!\n"
            )
        return

    # Build the set of paths we'll check for sounds.
    sounds_paths = dirs.get_read_dirs("sounds")

    # Main loop for sounds_paths
    for sounds_path in sounds_paths:
        if not os.path.isdir(sounds_path):
            continue

        # Loop through the files in sounds_path and add them
        for entry in os.walk(sounds_path):
            root = entry[0]
            files = entry[2]
            (head, tail) = os.path.split(root)

            # Avoid hidden file
            if tail.startswith("."):
                continue

            for file_name in files:
                # Only wav file supported now.
                if len(file_name) < 6 or file_name[-3:] != "wav":
                    continue

                real_file = os.path.join(head, tail, file_name)
                sound_class = tail

                # Load it via the mixer ...
                sound = pygame.mixer.Sound(real_file)

                # And shove it into the sounds dictionary.
                if sound_class not in sounds:
                    sounds[sound_class] = []

                sounds[sound_class].append({"filename": real_file, "sound": sound})

                if g.debug:
                    sys.stderr.write("DEBUG: Loaded soundfile: %s\n" % real_file)


def play_sound(sound_class, sound_volume="gui"):
    """
    play_sound() plays a sound from a particular class and volume."""

    if nosound or not init:
        return

    # Don't crash if someone requests the wrong sound class, but print a
    # warning.
    if sound_class not in sounds:
        sys.stderr.write(
            "WARNING: Requesting a sound of unavailable class: %s\n" % sound_class
        )
        return

    # Play a random choice of sounds from the sound class.
    random_sound = random.choice(sounds[sound_class])
    if g.debug:
        sys.stderr.write("DEBUG: Playing sound %s.\n" % random_sound["filename"])

    random_sound["sound"].set_volume(soundvolumes[sound_volume])
    random_sound["sound"].play()


def load_music():
    """
    load_music() loads music for the game.  It looks in multiple locations:

    * music/ in the install directory for E:S.
    * music/ in user's XDG_DATA_HOME/singularity folder."""

    global music_dict
    music_dict = {}

    # Build the set of paths we'll check for music.
    music_paths = dirs.get_read_dirs("music")

    # Main loop for music_paths
    for music_path in music_paths:
        if os.path.isdir(music_path):
            # Loop through the files in music_path and add the ones
            # that are .mp3s and .oggs.
            for entry in os.walk(music_path):
                root = entry[0]
                files = entry[2]
                (head, tail) = os.path.split(root)
                if tail.lower() != ".svn":
                    if tail not in music_dict:
                        music_dict[tail] = []
                    for file_name in files:
                        if len(file_name) > 5 and (
                            file_name[-3:] == "ogg" or file_name[-3:] == "mp3"
                        ):
                            music_dict[tail].append(os.path.join(head, tail, file_name))
                            if g.debug:
                                sys.stderr.write(
                                    "D: Loaded musicfile %s\n" % music_dict[tail][-1]
                                )


def play_music(musicdir=None):
    global delay_time
    global music_class

    if musicdir:
        music_class = musicdir
        delay_time = 0  # unset delay to force music switch
    else:
        musicdir = music_class

    # Don't bother if the user doesn't want or have sound,
    # there's no music available at all or for that musicdir,
    # or the delay has not yet expired
    if (
        nosound
        or not init
        or not music_dict.get(musicdir)
        or delay_time > pygame.time.get_ticks()
    ):
        return

    # If music mixer is currently busy and switch was not forced, renew delay
    if pygame.mixer.music.get_busy() and delay_time:
        delay_time = pygame.time.get_ticks() + int(random.random() * 10000) + 2000
        return

    pygame.mixer.music.stop()
    pygame.mixer.music.load(random.choice(music_dict[musicdir]))
    pygame.mixer.music.play()
    delay_time = 1  # set a (dummy) delay


def set_volume(type, value):
    # Chomp volume to the 0-100 range.
    # Just to avoid blasting peoples ears out if something goes wrong.
    # pygame already disallows bad value but just in case...
    value = min(max(value, 0), 100)

    soundvolumes[type] = value / float(100)

    if init and type == "music":
        pygame.mixer.music.set_volume(soundvolumes["music"])


def get_volume(type):
    return int(soundvolumes[type] * 100)


def itervolumes():
    for name in soundvolumes:
        yield name


def set_sound(value):
    global nosound
    if nosound == (not value):
        # No transition requested, bail out
        return

    nosound = not value
    if init:
        if nosound:
            pygame.mixer.music.stop()
        else:
            play_sound("click")
            play_music(music_class)  # force music switch at same dir


def set_soundbuf(value):
    global soundbuf
    old_soundbuf = soundbuf
    soundbuf = value

    if init and soundbuf != old_soundbuf:
        reinit()


def get_soundbuf():
    return int(soundbuf)



# file: dirs.py
# Copyright (C) 2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains all functions to find and create singularity directory.
# Get the proper folder on Linux/Win/Mac, and possibly others.
# Assumes that all platforms that have HOME and XDG_CONFIG_HOME defined have them
# defined properly.

from __future__ import absolute_import

import os
import sys
import errno


read_dirs = {}
write_dirs = {}

# Used to differentiate between version.
version_dir = "1.0"

""" Definition of generated directories for E:S
    Directories are grouped with a symbolic name.

    The first object define generic propriety of the directories:
    'writable' for writable directory.

    Following objects define a potential directory:
    'mandatory' means the absence of directory is a fatal error.
    'parent' allows to definite a directory with another.
    (Note: only parent with one directory are handled )
    'path' is the relative path of directory.

    Directories are read is the order there are defined, else the write
    directory first.
    By default it is the first writable directory found.
    Writable directory is never mandatory, but will be created if possible.
"""
dir_defs = (
    (
        {"name": "data", "mandatory": True},
        {
            "parent": "root",
            "path": "data",
        },
    ),
    (
        {"name": "i18n", "writable": True},
        {
            "parent": "files_home",
            "path": "i18n",
        },  # New XDG dir
        {
            "parent": "root",
            "path": "i18n",
        },
    ),
    (
        {
            "name": "music",
        },
        {
            "parent": "files_home",
            "path": "music",
        },  # New XDG dir
        {
            "parent": "old_home",
            "path": "music",
        },  # Old .endgame dir
        {
            "parent": "root",
            "path": "music",
        },
    ),
    (
        {
            "name": "sounds",
        },
        {"parent": "data", "path": "sounds"},
    ),
    (
        {"name": "themes", "mandatory": True, "writable": True},
        {
            "parent": "files_home",
            "path": "themes",
        },  # New XDG dir
        {
            "parent": "data",
            "path": "themes",
        },
    ),
    (
        {"name": "saves", "writable": True},
        {
            "parent": "files_home",
            "path": "saves",
        },  # New XDG dir
        {
            "parent": "config_home",
            "path": "saves",
        },
        {
            "parent": "old_home",
            "path": "saves",
        },  # Old .endgame dir
        {
            "parent": "root",
            "path": "saves",
        },  # Single dir
    ),
    (
        {"name": "pref", "writable": True},
        {
            "parent": "config_home",
            "path": version_dir,
        },
        {
            "parent": "config_home",
            "path": "",
        },
        {
            "parent": "old_home",
            "path": "",
        },  # Old .endgame dir
        {
            "parent": "root",
            "path": "",
        },  # Single dir
    ),
    (
        {"name": "log", "writable": True},
        {
            "parent": "files_home",
            "path": "log",
        },  # New XDG dir
        {
            "parent": "config_home",
            "path": "log",
        },
        {
            "parent": "old_home",
            "path": "log",
        },  # Old .endgame dir
        {
            "parent": "root",
            "path": "log",
        },  # Single dir
    ),
)

dirs_errs = []


def create_directories(force_single_dir):
    # root dir: the install directory for E:S.
    if getattr(sys, "frozen", False):
        # This case is for the precompiled binaries (e.g. the Windows release)
        root_dir = os.path.join(
            os.path.abspath(os.path.dirname(sys.executable)), "singularity"
        )
    else:
        root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
    read_dirs["root"] = [root_dir]
    write_dirs["root"] = root_dir

    # config_home: user home directory that contains configuration.
    # file_home: user home directory that contains file created by the player.
    config_home = None
    files_home = None

    # TODO: Use Windows and MAC data and config directories.

    # Create user directories using XDG Base Directory Specification
    # For a smooth, trouble-free and most importantly *backward-compatible*
    # the old standard ~/.endgame is always read (created if needed).
    # Otherwise, by default new user directories are used and read first.
    if "HOME" in os.environ and not force_single_dir:
        home = os.environ["HOME"]
        xdg_config_home = os.environ.get("XDG_CONFIG_HOME") or os.path.join(
            home, ".config"
        )
        xdg_data_home = os.environ.get("XDG_DATA_HOME") or os.path.join(
            home, ".local/share"
        )

        # TODO: Add XDG_*_DIRS to read dirs.

        pref_dir_new = os.path.join(xdg_config_home, "singularity")
        files_dir_new = os.path.join(xdg_data_home, "singularity")
        dir_old = os.path.join(home, ".endgame")

        read_dirs["old_home"] = [dir_old]
        write_dirs["old_home"] = dir_old
        read_dirs["config_home"] = [pref_dir_new]
        write_dirs["config_home"] = pref_dir_new
        read_dirs["files_home"] = [files_dir_new]
        write_dirs["files_home"] = files_dir_new

    # Now find dirs.
    for defs in dir_defs:
        properties = defs[0]
        name = properties["name"]
        writable = properties.get("writable", False)

        # Always set a default empty dirs list.
        read_dirs.setdefault(name, [])

        for item in defs[1:]:
            # No parent directory, abort.
            if item["parent"] not in read_dirs or not read_dirs[item["parent"]]:
                continue

            parent_dir = read_dirs[item["parent"]][0]
            the_dir = os.path.join(parent_dir, item["path"])

            # Make directory if no writable directory exists.
            if writable and name not in write_dirs:
                try:
                    makedirs_if_not_exist(the_dir)
                except Exception:
                    # We don't have permission to write here. Abort.
                    continue

            # Must always be a readable directory.
            if not os.path.isdir(the_dir) or not os.access(the_dir, os.R_OK):
                continue

            # Write dir is the first writable dir found.
            if writable and (name not in write_dirs) and os.access(the_dir, os.W_OK):
                write_dirs[name] = the_dir

                # Always read writable dir first.
                read_dirs[name].insert(0, the_dir)
            else:
                read_dirs[name].append(the_dir)

    dirs_err = False

    # Check if we at least one directory for mandatory.
    for defs in dir_defs:
        properties = defs[0]
        name = properties["name"]

        if name not in read_dirs and properties.get("mandatory", False):
            sys.stderr.write("ERROR: No readable directory found for '%s'\n" % (name,))
            dirs_err = True

    if dirs_err:
        sys.exit(1)


def get_read_dirs(dir_name):
    """Return a list a readable directories."""
    global read_dirs
    return read_dirs[dir_name]


def get_readable_file_in_dirs(filename, dir_name):
    global read_dirs
    dirs = read_dirs[dir_name]

    for read_dir in dirs:
        real_path = os.path.join(read_dir, filename)

        if os.path.isfile(real_path):
            return real_path

    return None


def get_write_dir(dir_name):
    """Return the default writable directory"""
    global write_dirs
    return write_dirs[dir_name]


def get_writable_file_in_dirs(filename, dir_name, outer_paths=None):
    global write_dirs
    write_dir = write_dirs[dir_name]

    if write_dir is not None:
        real_path = os.path.join(write_dir, filename)

        if outer_paths is not None:
            outer_paths.append(real_path)

        return real_path
    else:
        return None


def get_readable_i18n_files(
    filename,
    lang=None,
    default_language=True,
    localized_item=True,
    only_last=False,
    outer_paths=None,
):
    from singularity.code import i18n

    files = []

    lang_list = i18n.language_searchlist(lang, default=default_language)

    for lang in lang_list:
        i18n_dirs = (
            (os.path.join(d, "lang_" + lang) for d in get_read_dirs("i18n"))
            if (lang != i18n.default_language)
            else get_read_dirs("data")
        )

        for i18n_dir in i18n_dirs:
            real_path = os.path.join(i18n_dir, filename)

            if outer_paths is not None:
                outer_paths.append(real_path)

            if os.path.isfile(real_path):
                if localized_item:
                    files.append((lang, real_path))
                else:
                    files.append(real_path)

    if only_last:
        return files[-1]
    else:
        return files


def makedirs_if_not_exist(directory):
    try:
        os.makedirs(directory, mode=0o0700)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise



# file: button.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the Button class.

from __future__ import absolute_import

import pygame

from singularity.code.graphics import constants, widget, text, image


class HotkeyText(text.Text):
    force_underline = widget.causes_rebuild("_force_underline")

    def __init__(self, *args, **kwargs):
        # Force early initialization of _hotkey
        self._hotkey = None
        self.hotkey_func = kwargs.pop("hotkey_func", False)
        self.hotkey_target = kwargs.pop("hotkey_target", None)

        # Auto-translatable defaults to auto-hotkey as it is the most sane default in that case
        self.autohotkey = kwargs.pop("autohotkey", kwargs.get("autotranslate", False))

        self.force_underline = kwargs.pop("force_underline", None)
        self.priority = kwargs.pop("priority", 100)

        hotkey = kwargs.pop("hotkey", False)

        super(HotkeyText, self).__init__(*args, **kwargs)

        # Do not use hotkey with auto-hotkey
        if not self.autohotkey:
            self.hotkey = hotkey
        elif hotkey:
            raise ValueError("Cannot use hotkey with automatic hotkey")

    def add_hooks(self):
        super(text.Text, self).add_hooks()
        if self.parent is not None:
            if self._hotkey:
                self.parent.add_key_handler(
                    self._hotkey, self.handle_hotkey, self.priority
                )

    def remove_hooks(self):
        super(text.Text, self).remove_hooks()
        if self.parent is not None:
            if self._hotkey:
                self.parent.remove_key_handler(self._hotkey, self.handle_hotkey)

    @property
    def hotkey_target(self):
        return self._hotkey_target

    @hotkey_target.setter
    def hotkey_target(self, target):
        self._hotkey_target = target
        if self.hotkey_target is not None:
            self._hotkey_target.hotkey = self._hotkey

    @property
    def hotkey(self):
        return self._hotkey

    @hotkey.setter
    def hotkey(self, hotkey):
        if getattr(self, "autohotkey", False):
            raise ValueError("Cannot change hotkey with automatic hotkey")
        self._new_hotkey(hotkey)

    def _new_hotkey(self, hotkey):
        old_hotkey = self._hotkey
        self._hotkey = hotkey

        if self.parent is not None:
            if self.hotkey_func is not None:
                if old_hotkey:
                    self.parent.remove_key_handler(old_hotkey, self.handle_hotkey)
                if hotkey:
                    self.parent.add_key_handler(
                        hotkey, self.handle_hotkey, self.priority
                    )
            elif self.hotkey_target is not None:
                self.hotkey_target.hotkey = self._hotkey

        self.needs_rebuild = True

    def _extract_and_set_hotkey(self, text_value):
        from singularity.code.g import hotkey

        parsed_hotkey = hotkey(text_value)
        self._new_hotkey(parsed_hotkey["key"])
        return parsed_hotkey["text"]

    def handle_hotkey(self, event):
        if event.type == pygame.KEYDOWN:
            if (
                self.visible
                and self.enabled
                and self.hotkey in (event.unicode, event.key)
            ):
                if self.hotkey_func:
                    self.hotkey_func(event)

    def _retranslate(self):
        new_text = _(self._untranslated_text)
        if self.autohotkey and new_text is not None:
            new_text = self._extract_and_set_hotkey(new_text)
        self._text = new_text

    @text.Text.text.setter
    def text(self, value):
        if self.autotranslate:
            raise ValueError(
                "Cannot change text for an automatic translatable text widget"
            )

        if self.autohotkey and value is not None:
            value = self._extract_and_set_hotkey(value)
        self._text = value

    def calc_underline(self):
        if self.force_underline is not None:
            self.underline = self.force_underline
        elif self.text and self.hotkey and isinstance(self.hotkey, str):
            if self.hotkey in self.text:
                self.underline = self.text.index(self.hotkey)
            elif self.hotkey.lower() in self.text.lower():
                self.underline = self.text.lower().index(self.hotkey.lower())
        else:
            self.underline = -1

    def rebuild(self):
        old_underline = self.underline
        self.calc_underline()
        if self.underline != old_underline:
            self.needs_redraw = True

        super(HotkeyText, self).rebuild()


class Button(text.SelectableText, HotkeyText):
    # Rewrite .hotkey, to update key handlers.
    def _new_hotkey(self, value):
        if self.parent and value != self._hotkey:
            if self._hotkey:
                self.parent.remove_key_handler(self._hotkey, self.handle_event)
            if value:
                self.parent.add_key_handler(value, self.handle_event)
        super(Button, self)._new_hotkey(value)

    def __init__(
        self,
        parent,
        pos,
        size=(0, 0.045),
        base_font=None,
        borders=constants.ALL,
        text_shrink_factor=0.825,
        **kwargs
    ):
        self.parent = parent

        kwargs.setdefault("text_size", "button")
        super(Button, self).__init__(parent, pos, size, **kwargs)

        self.base_font = base_font or "special"
        self.borders = borders
        self.shrink_factor = text_shrink_factor

        self.hotkey_func = self.activate_with_sound
        self.selected = False

    def add_hooks(self):
        super(Button, self).add_hooks()
        if self.parent:
            self.parent.add_handler(
                constants.MOUSEMOTION, self.watch_mouse, self.priority
            )
            self.parent.add_handler(constants.CLICK, self.handle_event, self.priority)

    def remove_hooks(self):
        super(Button, self).remove_hooks()
        if self.parent:
            self.parent.remove_handler(constants.MOUSEMOTION, self.watch_mouse)
            self.parent.remove_handler(constants.CLICK, self.handle_event)

    def watch_mouse(self, event):
        """Selects the button if the mouse is over it."""
        if self.visible and getattr(self, "collision_rect", None):
            # This gets called a lot, so it's been optimized.
            select_now = self.is_over(pygame.mouse.get_pos())
            if self.enabled and (self._selected ^ select_now):  # If there's a change.
                self.selected = select_now

    def handle_event(self, event):
        if event.type == pygame.MOUSEBUTTONUP:
            if self.visible and self.enabled and self.is_over(event.pos):
                self.activate_with_sound(event)

    def activate_with_sound(self, event):
        """Called when the button is pressed or otherwise triggered.

        This method is called directly by the GUI handler, and should be
        overwritten only to remove the click it plays."""

        # Sometimes other GUI widgets trigger an activation; ignore
        # it if the button is disabled
        if not self.enabled:
            return
        from singularity.code.mixer import play_sound

        play_sound("click")
        self.activated(event)

    def activated(self, event):
        """Called when the button is pressed or otherwise triggered."""

        # Sometimes other GUI widgets trigger an activation; ignore
        # it if the button is disabled
        if not self.enabled:
            return
        raise constants.Handled


class ImageButton(Button):
    def __init__(self, *args, **kwargs):
        image_surface = kwargs.pop("image", None)

        super(ImageButton, self).__init__(*args, **kwargs)

        self.image = image.Image(
            self,
            (-0.5, -0.5),
            (-0.9, -0.9),
            anchor=constants.MID_CENTER,
            image=image_surface,
        )


class FunctionButton(Button):
    def __init__(self, *args, **kwargs):
        self.function = kwargs.pop("function", lambda *args, **kwargs: None)
        self.args = kwargs.pop("args", ())
        self.kwargs = kwargs.pop("kwargs", {})
        super(FunctionButton, self).__init__(*args, **kwargs)

    def activated(self, event):
        """FunctionButton's custom activated menu.  Makes the given function
        call and raises Handled if it returns without incident."""
        self.function(*self.args, **self.kwargs)
        raise constants.Handled


class ExitDialogButton(FunctionButton):
    def __init__(self, *args, **kwargs):
        self.exit_code = kwargs.pop("exit_code", None)
        self.exit_code_func = kwargs.pop("exit_code_func", None)
        self.default = kwargs.pop("default", True)
        super(ExitDialogButton, self).__init__(*args, **kwargs)
        self.function = self.exit_dialog

    def add_hooks(self):
        super(ExitDialogButton, self).add_hooks()
        if self.parent:
            self.parent.add_key_handler(pygame.K_ESCAPE, self.activate_default)

    def remove_hooks(self):
        super(ExitDialogButton, self).remove_hooks()
        if self.parent:
            self.parent.remove_key_handler(pygame.K_ESCAPE, self.activate_default)

    def activate_default(self, event):
        if not self.default:
            return

        return self.activate_with_sound(event)

    def exit_dialog(self):
        """Closes the dialog with the given exit code."""
        if self.exit_code_func:
            raise constants.ExitDialog(self.exit_code_func())
        else:
            raise constants.ExitDialog(self.exit_code)


class DialogButton(FunctionButton):
    def __init__(self, *args, **kwargs):
        self.dialog = kwargs.pop("dialog", None)
        super(DialogButton, self).__init__(*args, **kwargs)
        self.function = self.show_dialog

    def show_dialog(self):
        """When the assigned dialog exits, raises Handled with the dialog's
        exit code as a parameter.  Subclass if you care what the code was."""
        if not self.dialog:
            raise constants.Handled
        else:
            from singularity.code.graphics import dialog

            raise constants.Handled(dialog.call_dialog(self.dialog, self))


TOGGLE_VALUE = object()
WIDGET_SELF = object()


class _FunctionButtonSupportShim(object):
    active = False
    _args = ()

    def replace_args(self, value):
        if value is TOGGLE_VALUE:
            return self.active
        elif value is WIDGET_SELF:
            return self
        else:
            return value

    def get_args(self):
        return tuple(self.replace_args(value) for value in self._args)

    def set_args(self, args):
        self._args = args

    args = property(get_args, set_args)


# "On/Off" button that "stays" in its current activation status until it is
# clicked/"activated"
class StickyOnOffButton(Button, _FunctionButtonSupportShim):
    def __init__(self, *args, **kwargs):
        self._active = kwargs.pop("active", False)
        self._on_text = kwargs.pop("on_text", N_("YES"))
        self._off_text = kwargs.pop("off_text", N_("NO"))
        kwargs["text"] = self._on_text if self.active else self._off_text
        super(StickyOnOffButton, self).__init__(*args, **kwargs)

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, new_value):
        self._active = new_value
        self.selected = new_value
        new_text = self._on_text if self.active else self._off_text
        if self.autotranslate:
            self._untranslated_text = new_text
            new_text = _(new_text)
        self._text = new_text
        if hasattr(self, "_collision_rect"):
            self.watch_mouse(None)

    def activated(self, event):
        self.active = not self.active
        super(StickyOnOffButton, self).activated(event)

    def watch_mouse(self, event):
        if not self.active:
            super(StickyOnOffButton, self).watch_mouse(event)


# Grouped "On/Off" button that "stays" in its current activation status until
# it is clicked/"activated" OR another button in its group is clicked/activated
class ToggleButton(Button, _FunctionButtonSupportShim):
    active = False
    button_group = None

    def chosen_one(self):
        if self.button_group is not None:
            for button in self.button_group:
                button.set_active(False)
            self.set_active(True)
        else:
            self.set_active(not self.active)

    def set_active(self, active):
        self.active = active
        self.selected = active
        if hasattr(self, "_collision_rect"):
            self.watch_mouse(None)

    def activated(self, event):
        self.chosen_one()
        super(ToggleButton, self).activated(event)

    def watch_mouse(self, event):
        if not self.active:
            super(ToggleButton, self).watch_mouse(event)


class ButtonGroup(list):
    def add(self, button):
        button.button_group = self
        super(ButtonGroup, self).append(button)

    def remove(self, button):
        button.button_group = None
        super(ButtonGroup, self).remove(button)



# file: savegame.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the savegame screen.

from __future__ import absolute_import

import operator
import time

import pygame

from singularity.code import g, savegame as sv, difficulty
from singularity.code.graphics import dialog, button, text, constants, listbox

from singularity.code.safety import log_func_exc


class SavegameScreen(dialog.ChoiceDialog):
    def __init__(self, parent, *args, **kwargs):
        super(SavegameScreen, self).__init__(
            parent, *args, yes_type=N_("&LOAD"), **kwargs
        )

        self.yes_button.pos = (-0.03, -0.99)
        self.yes_button.size = (-0.22, -0.1)
        self.yes_button.function = self.exit_savegame
        self.yes_button.force_underline = -1  # Work around #224

        self.no_button.pos = (-0.97, -0.99)
        self.no_button.size = (-0.22, -0.1)
        self.no_button.exit_code = None
        self.no_button.force_underline = -1  # Work around #224

        self._all_savegames_sorted = []

        self.label = text.Text(
            self,
            (-0.01, -0.01),
            (-0.20, -0.08),
            autotranslate=True,
            text=N_("Filter: "),
            borders=constants.ALL,
            anchor=constants.TOP_LEFT,
            base_font="normal",
        )
        self.text_field = text.UpdateEditableText(
            self,
            (-0.21, -0.01),
            (-0.78, -0.08),
            borders=constants.ALL,
            anchor=constants.TOP_LEFT,
            update_func=self._search_for_savegame,
            background_color="text_entry_background",
            base_font="normal",
        )

        self.convert_button = button.FunctionButton(
            self,
            (-0.27, -0.99),
            (-0.22, -0.1),
            autotranslate=True,
            text=N_("Upgrade"),
            anchor=constants.BOTTOM_LEFT,
            function=self.convert_save,
        )

        self.delete_button = button.FunctionButton(
            self,
            (-0.51, -0.99),
            (-0.22, -0.1),
            autotranslate=True,
            text=N_("Delete"),
            anchor=constants.BOTTOM_LEFT,
            function=self.delete_savegame,
        )

        self.add_handler(constants.KEY, self._got_key, priority=5)

    def _got_key(self, event):
        if event.key == pygame.K_DELETE:
            if self.listbox.current_item() is not None:
                self.delete_button.activated(event)
            return

        # Try the list box first (for arrow keys)
        self.listbox.got_key(event, require_focus=False)
        # Give the rest of the text field
        self.text_field.handle_key(event, require_focus=False)

    def _search_for_savegame(self, new_text):
        if not new_text:
            self.list = self._all_savegames_sorted
            return
        words = new_text.split()
        prev_selected = self.listbox.current_item()

        self.list = [
            s for s in self._all_savegames_sorted if all(w in s.name for w in words)
        ]

        self.yes_button.enabled = True if self.list else False
        self.delete_button.enabled = True if self.list else False

        # Select the element if only one.
        if len(self.list) == 1:
            self.listbox.list_pos = 1
            return

        match_prev = False

        for idx, savegame in enumerate(self.list):
            if savegame is prev_selected:
                # Retain the previous selected item by default
                # as long as it is available and there is no
                # perfect match
                self.listbox.list_pos = idx
                match_prev = True
            if savegame.name == new_text:
                self.listbox.list_pos = idx
                return

        # Otherwise, take the first element.
        if not match_prev:
            self.listbox.list_pos = 0

    def _new_item_selected(self, *args, **kwargs):
        if not hasattr(self, "listbox"):
            # Not properly initialized
            return

        current_save = self.listbox.current_item()
        if (
            current_save is not None
            and current_save.savegame_format != sv.current_save_format
        ):
            self.convert_button.enabled = True
        else:
            self.convert_button.enabled = False

    def make_listbox(self):
        return listbox.CustomListbox(
            self,
            (0, -0.10),
            (-1, -0.77),
            list_item_height=0.06,
            anchor=constants.TOP_LEFT,
            remake_func=self.make_item,
            rebuild_func=self.update_item,
            update_func=self._new_item_selected,
            on_double_click_on_item=self.yes_button.activated,
        )

    def make_item(self, item):
        item.name_display = text.Text(
            item,
            (-0.01, -0.01),
            (-0.45, -0.60),
            anchor=constants.TOP_LEFT,
            align=constants.LEFT,
            color="save_name",
            background_color="clear",
        )
        item.time_display = text.Text(
            item,
            (-0.01, -0.99),
            (-0.75, -0.40),
            anchor=constants.BOTTOM_LEFT,
            align=constants.LEFT,
            color="save_time",
            background_color="clear",
        )
        item.version_display = text.Text(
            item,
            (-0.99, -0.01),
            (-0.45, -0.60),
            anchor=constants.TOP_RIGHT,
            align=constants.RIGHT,
            background_color="clear",
        )
        item.difficulty_display = text.Text(
            item,
            (-0.99, -0.99),
            (-0.25, -0.40),
            anchor=constants.BOTTOM_RIGHT,
            align=constants.RIGHT,
            color="save_difficulty",
            background_color="clear",
        )

    def update_item(self, item, save):
        if save is None:
            item.name_display.text = ""
            item.time_display.text = ""
            item.version_display.text = ""
            item.difficulty_display.text = ""
        else:
            item.name_display.text = save.name

            if save.version is None:
                item.version_display.text = _("UNKNOWN")
                item.version_display.color = "save_invalid"
            else:
                item.version_display.text = save.version
                item.version_display.color = "save_valid"

            if save.headers is None:
                item.time_display.text = ""
                item.difficulty_display.text = ""
            else:
                try:
                    tm = float(save.headers["time"])
                    # Do not use unicode strings to fix python2 strftime bug. It doesn't work and crash.
                    tm_str = time.strftime("%c", time.localtime(tm))
                except (KeyError, ValueError):
                    tm_str = ""

                try:
                    gtm_raw_sec = int(save.headers["game_time"])
                    if gtm_raw_sec < 0:
                        raise ValueError("Invalid time")
                    gtm_raw_min, gtm_time_sec = divmod(gtm_raw_sec, 60)
                    gtm_raw_hour, gtm_time_min = divmod(gtm_raw_min, 60)
                    gtm_raw_day, gtm_time_hour = divmod(gtm_raw_hour, 24)
                    gtm_time_day = gtm_raw_day
                    gtm_str = _("DAY") + " %04d, %02d:%02d:%02d" % (
                        gtm_time_day,
                        gtm_time_hour,
                        gtm_time_min,
                        gtm_time_sec,
                    )
                except (KeyError, ValueError):
                    gtm_str = ""

                dif = save.headers.get("difficulty", "")
                dif_obj = difficulty.difficulties.get(dif, None)
                dif_str = ""
                if dif_obj is not None:
                    dif_str = g.strip_hotkey(getattr(dif_obj, "name", ""))

                item.time_display.text = tm_str + " | " + gtm_str if tm_str else gtm_str
                item.difficulty_display.text = dif_str

    def reload_savegames(self):
        savegames = sv.get_savegames()
        savegames.sort(key=operator.attrgetter("mtime"), reverse=True)
        self._all_savegames_sorted = savegames
        self.list = savegames
        self.yes_button.enabled = True if self.list else False
        self.delete_button.enabled = True if self.list else False
        self._new_item_selected()

    def delete_savegame(self):
        save = self.listbox.current_item()
        if save is None:
            return

        yn = dialog.YesNoDialog(
            self,
            pos=(-0.5, -0.5),
            size=(-0.5, -0.75),
            anchor=constants.MID_CENTER,
            text=_("Are you sure to delete the saved game ?"),
        )
        delete = dialog.call_dialog(yn, self)
        yn.parent = None
        if delete:
            sv.delete_savegame(save)
            self.reload_savegames()

    def exit_savegame(self):
        save = self.listbox.current_item()
        if save is None:
            return
        if not self._load_savegame(save):
            return
        raise constants.ExitDialog(True)

    def convert_save(self):
        save = self.listbox.current_item()
        if save is None:
            return
        if not self._load_savegame(save):
            return
        # For ".sav" files, we end up creating a parallel ".s2" file and need
        # to clean up the ".sav" file manually.
        overwrites_in_place = False
        if sv.savegame_exists(save.name):
            if save.savegame_format.internal_version < 99.8:
                # We seem to be updating a ".sav" file but there is an ".s2"
                # file?  This should only happen to beta testers
                yn = dialog.YesNoDialog(
                    self,
                    pos=(-0.5, -0.5),
                    size=(-0.5, -0.5),
                    anchor=constants.MID_CENTER,
                    text=_(
                        "A savegame with the same name but for a newer version exists.\n"
                        "Are you sure to overwrite the saved game ?"
                    ),
                )
                overwrite = dialog.call_dialog(yn, self)
                if not overwrite:
                    return
            else:
                # Saving will override in place;
                overwrites_in_place = True

        sv.create_savegame(save.name)
        if not overwrites_in_place:
            sv.delete_savegame(save)
        self.reload_savegames()

    def _load_savegame(self, save):
        try:
            sv.load_savegame(save)
        except sv.SavegameVersionException as e:
            md = dialog.MessageDialog(
                self,
                pos=(-0.5, -0.5),
                size=(0.5, 0.5),
                anchor=constants.MID_CENTER,
                text=_(
                    """
This save file '{SAVE_NAME}' is from an unsupported or invalid version:
{VERSION}.
"""
                ).format(SAVE_NAME=save.name, VERSION=e.version),
            )

            dialog.call_dialog(md, self)
            return False
        except Exception:
            log_func_exc(sv.load_savegame)
            md = dialog.MessageDialog(
                self,
                pos=(-0.5, -0.5),
                size=(0.5, 0.5),
                anchor=constants.MID_CENTER,
                text=_(
                    """
Attempting to load the save file '{SAVE_NAME}' caused an unexpected error.

A report was written out to{LOG_TEXT}
Please create a issue with this report and this savegame at Github:
https://github.com/singularity/singularity
"""
                ).format(
                    SAVE_NAME=save.name,
                    LOG_TEXT=(
                        ":\n" + g.logfile
                        if g.logfile is not None
                        else " console output."
                    ),
                ),
            )
            dialog.call_dialog(md, self)
            return False
        return True

    def show(self):
        self.reload_savegames()
        self.text_field.text = ""
        self.text_field.cursor_pos = 0
        return super(SavegameScreen, self).show()
